!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! adding the identity to a dense matrix
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine cqrm_dsmat_addi(a, v, prio, info)

  use cqrm_dsmat_mod, alias=>cqrm_dsmat_addi
  use qrm_dscr_mod

  implicit none

  type(cqrm_dsmat_type), target   :: a
  integer, optional               :: prio
  complex(r32), optional          :: v
  integer, optional               :: info
                                   
  type(qrm_dscr_type)             :: qrm_dscr
                                   
  ! error handling                 
  integer                         :: err
  character(len=*), parameter     :: name='qrm_dsmat_addi'

  err = 0

  if(.not.a%inited) then
     err = 1000
     call qrm_error_print(err, name)
     goto 9999
  end if

  call qrm_dscr_init(qrm_dscr, nocuda=.true.)
  call cqrm_dsmat_addi_async(qrm_dscr, a, v, prio)
  call qrm_barrier(qrm_dscr)
  err = qrm_dscr%info

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  if(present(info)) info = err
  return

end subroutine cqrm_dsmat_addi

subroutine cqrm_dsmat_addi_async(qrm_dscr, a, v, prio)
  use cqrm_dsmat_mod, s=>cqrm_dsmat_addi_async
  use qrm_dscr_mod
  use cqrm_dsmat_task_mod
  use qrm_string_mod
  use qrm_parameters_mod
#if defined (have_starpu)
  use cqrm_starpu_mod
  use fstarpu_mod
#endif
  
  implicit none

  type(qrm_dscr_type)            :: qrm_dscr
  type(cqrm_dsmat_type), target  :: a
  integer, optional              :: prio
  complex(r32), optional         :: v 

  type(cqrm_block_type), pointer :: aij
  integer                        :: mb, nb, iprio
  integer                        :: ii, ij, im, in, il, mm, nn
  complex(r32)                   :: iv
  integer                        :: first_block
  integer                        :: last_block
  integer                        :: b, bi, bj, bm, bn, bl
  
  ! error handling
  integer                        :: info, err
  character(len=*), parameter    :: name='qrm_dsmat_addi_async'


  if(qrm_dscr%info.ne.0) return
  info = 0

  if(.not.a%inited) then
     info = 1000
     call qrm_error_print(info, name)
     goto 9999
  end if

  __QRM_IFPRESENT3(iprio, prio, 0)
  __QRM_IFPRESENT3(iv, v, qrm_cone)

  first_block = 1
  last_block  = cqrm_dsmat_inblock(a, min(a%m,a%n))

  do b = first_block, last_block
     call cqrm_dsmat_block_ijmnl(a, 1, 1, a%m, a%n, 0, b, b, bi, bj, bm, bn, bl)
     if(min(bm,bn).gt.0) then
        aij => a%blocks(b,b)
        call cqrm_block_addi_task(qrm_dscr, &
             aij,                           &
             bi, bj,                        &
             bm, bn,                        &
             iv, iprio)
     end if
  end do
  
9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return

end subroutine cqrm_dsmat_addi_async
