!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT, Inria
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! computing a triangular system solution
!!
!! ##############################################################################################

#include "qrm_common.h"


subroutine dqrm_dsmat_sytrf(uplo, a, m, k, prio, info)

  use dqrm_dsmat_mod, alias => dqrm_dsmat_sytrf
  use qrm_dscr_mod
  use qrm_parameters_mod
  implicit none

  type(dqrm_dsmat_type)       :: a
  character                   :: uplo
  integer, optional           :: m, k, prio, info

  type(qrm_dscr_type)         :: qrm_dscr

  ! error handling
  integer                     :: err
  character(len=*), parameter :: name='qrm_dsmat_sytrf'

  err = 0

  call qrm_dscr_init(qrm_dscr)

  call dqrm_dsmat_sytrf_async(qrm_dscr, uplo, a, m, k, prio)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, info)

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  return

end subroutine dqrm_dsmat_sytrf

subroutine dqrm_dsmat_sytrf_async(qrm_dscr, uplo, a, m, k, prio)

  use dqrm_dsmat_mod, alias => dqrm_dsmat_sytrf_async
  use qrm_parameters_mod
  use qrm_dscr_mod
  use dqrm_dsmat_task_mod
  implicit none

  type(qrm_dscr_type)         :: qrm_dscr
  type(dqrm_dsmat_type)       :: a
  character                   :: uplo
  integer, optional           :: prio, m, k

  integer                     :: i, j, l
  integer                     :: im, in, ik, mm, nn, kk
  integer                     :: last_mblock, last_nblock, last_kblock

  ! error handling
  integer                     :: info
  character(len=*), parameter :: name='qrm_dsmat_sytrf_async'

  if(qrm_dscr%info.ne.0) return
  if(min(a%m,a%n).le.0) return
  info = 0

  __QRM_IFPRESENT3(im, m, a%m)
  __QRM_IFPRESENT3(ik, k, a%m)

  ik = min(im,ik)
  if(ik.le.0) return
  
  last_mblock = dqrm_dsmat_inblock(a, im)
  last_kblock = dqrm_dsmat_inblock(a, ik)
  
  if(uplo.eq.qrm_upper) then
     do l=1, last_kblock
        if(l.eq.last_kblock) then
           kk = ik-a%f(l)+1
           mm = min(a%f(l+1)-a%f(l), im-a%f(l)+1)
        else
           kk = a%f(l+1)-a%f(l)
           mm = a%f(l+1)-a%f(l)
        end if
              
        call dqrm_sytrf_task(qrm_dscr,             &
             uplo,                                 & 
             mm, kk,                               &
             a%blocks(l,l),                        & 
             prio )
        do j = l+1, last_mblock
           if(j.eq.last_mblock) then
              nn = im-a%f(j)+1
           else
              nn = a%f(j+1)-a%f(j)
           end if
           call dqrm_trsm_task(qrm_dscr,           & 
                qrm_left, uplo,                    &
                qrm_transp, qrm_no_unit,           &
                kk, nn, mm,                        &
                qrm_done, a%blocks(l,l),           &
                a%blocks(l,j), prio)

           do i = l+1, j-1
              call dqrm_gemm_task(qrm_dscr,        &
                   qrm_transp, qrm_no_transp,      &
                   a%f(i+1)-a%f(i), nn, kk,        &
                   qrm_dmone,                      &
                   a%blocks(l,i), a%blocks(l,j),   &
                   qrm_done, a%blocks(i,j), prio)
           end do
           call dqrm_syrk_task(qrm_dscr,           &
                uplo, qrm_transp,                  &
                nn, kk,                            &
                qrm_dmone, a%blocks(l,j),          &
                qrm_done,  a%blocks(j,j), prio)
        end do
     end do
  else
  end if

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return

end subroutine dqrm_dsmat_sytrf_async
