!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! applying Householder transformations to a matrix
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine sqrm_dsmat_gemqr(trans, a, t, x, ib, bh, work, m, n, k, prio, info)

  use sqrm_dsmat_mod, alias=>sqrm_dsmat_gemqr
  use qrm_dscr_mod

  implicit none

  character                   :: trans
  type(sqrm_dsmat_type)       :: a, t, x
  type(sqrm_ws_type)          :: work

  integer                     :: ib, bh
  integer, optional           :: m, n, k, prio
  integer, optional           :: info

  type(qrm_dscr_type)         :: qrm_dscr
  
  ! error handling
  integer                     :: err
  character(len=*), parameter :: name='qrm_dsmat_gemqr'

  call qrm_dscr_init(qrm_dscr)

  call sqrm_dsmat_gemqr_async(qrm_dscr, trans, a, t, x, ib, bh, work, m, n, k, prio)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, info)

  call qrm_dscr_destroy(qrm_dscr)

  return
end subroutine sqrm_dsmat_gemqr

subroutine sqrm_dsmat_gemqr_async(qrm_dscr, trans, a, t, x, ib, bh, work, m, n, k, prio)

  use sqrm_dsmat_mod, alias=>sqrm_dsmat_gemqr_async
  use qrm_dscr_mod
  use sqrm_dsmat_task_mod

  implicit none

  type(qrm_dscr_type)         :: qrm_dscr
  character                   :: trans
  type(sqrm_dsmat_type)       :: a, t, x
  type(sqrm_ws_type)          :: work
  integer                     :: ib, bh
  integer, optional           :: m, n, k, prio

  integer                     :: i, j, p, s, l, nbr, nbc, nbcx, nbe, first, ibh
  integer                     :: mm, nn, kk, ll, im, in, ik, nk, ki
  integer, allocatable        :: ks(:)
  
  ! error handling
  integer                     :: info
  character(len=*), parameter :: name='qrm_dsmat_gemqr_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  __QRM_IFPRESENT3(im, m, a%m)
  __QRM_IFPRESENT3(in, n, x%n)
  __QRM_IFPRESENT3(ik, k, a%n)

  ik = min(im,ik)

  if(min(ik, in).le.0) return
  
  nbr  = sqrm_dsmat_inblock(a, im) ! number of block-rows
  nbc  = sqrm_dsmat_inblock(a, ik) ! number of block-cols
  nbe  = min(nbr,nbc)
  nbcx = sqrm_dsmat_inblock(x, in)

  if(bh.le.0) then
     ibh = nbr
  else
     ibh = bh
  end if

  allocate(ks(a%nbr+1))
  

  if(trans .eq. qrm_transp) then
     do p=1, nbe
        kk  = min(ik-a%f(p)+1,a%f(p+1)-a%f(p))
        ki  = p
        nk  = 0
        
        do l = p, nbr
           if(.not.qrm_allocated(a%blocks(l,p)%c)) exit
        
           mm = min(a%f(l+1)-a%f(l), im-a%f(l)+1)           ! this block is a killer if:
           if(  l.eq.p        .or.                        & ! it's the diagonal block or
                (l-ki.ge.ibh)  .and.                      & ! it's at least bh below the last killer
                (mm.ge.kk) .and.                          & ! and it has more rows than columns
                .not. qrm_allocated(a%blocks(l,p)%stair) )& ! and does not have a staircase structure
             then

              nk = nk+1
              ks(nk)=l
              ki=l
           end if
        end do
        ks(nk+1)=l

        ! the list of killers is now ready. go through it
        ! write(*,'(" ")')
        do ki=1, nk
           i = ks(ki)

           mm = min(a%f(i+1)-a%f(i), im-a%f(i)+1)
           do j=1, nbcx
              nn = min(in-x%f(j)+1,x%f(j+1)-x%f(j))
              ! write(*,'("geqrt  :",i2,x,i2,"  mm:",i3," kk:",i3," nn:",i3)')i,p,mm,kk,nn
              if(qrm_allocated(t%blocks(i,p)%c))       &
              call sqrm_higemqrt_task(qrm_dscr,        &
                   qrm_transp,                    &
                   mm, nn, kk, max(nn,kk), ib,         &
                   1, 1,                               &
                   a%blocks(i,p),                      &
                   t%blocks(i,p),                      &
                   x%blocks(i,j),                      &
                   work, prio)               
           end do                                       

           do l=i+1,ks(ki+1)-1
              mm = min(im-a%f(l)+1,a%f(l+1)-a%f(l))
              ll = 0
              do j=1, nbcx                              
                 nn = min(in-x%f(j)+1,x%f(j+1)-x%f(j))
                 ! write(*,'("  tpmqrt s:",i2,x,i2,"->",i2,x,i2,"  mm:",i3," kk:",i3," nn:",i3)')i,p,l,p,mm,kk,nn
                 ! write(*,*)'col',p,'killing ',l, 'using',i
                 if(qrm_allocated(a%blocks(l,p)%c))    &
                      call sqrm_hitpmqrt_task(qrm_dscr,     &
                           qrm_transp,                 &
                           mm, nn, kk, ll, max(nn,kk), ib,  &
                           1, 1,                            &
                           a%blocks(l,p),                   &
                           t%blocks(l,p),                   &
                           x%blocks(i,j),                   &
                           x%blocks(l,j),                   &
                           work, prio)
              end do
           end do

           if(i.ne.p) then
              do j=1, nbcx
                 nn = min(in-x%f(j)+1,x%f(j+1)-x%f(j))
                 ! write(*,'("tpmqrt t:",i2,x,i2,"->",i2,x,i2,"  kk:",i3," nn:",i3)')p,p,i,p,kk,nn
                 ! write(*,*)'col',p,'killing ',i, 'using',p,qrm_allocated(t%blocks(i,nbe+p)%c)
                 if(qrm_allocated(a%blocks(i,p)%c))        &
                      call sqrm_hitpmqrt_task(qrm_dscr,    &
                           qrm_transp,                &
                           kk, nn, kk, kk, max(kk,nn), ib, &
                           1, 1,                           &
                           a%blocks(i,p),                  &
                           t%blocks(i,nbe+p),              &
                           x%blocks(p,j),                  &
                           x%blocks(i,j),                  &
                           work, prio)
              end do
           end if

        end do

     end do
  else


     do p=nbe, 1, -1
        kk  = min(ik-a%f(p)+1,a%f(p+1)-a%f(p))
        ki  = p
        nk  = 0
        
        do l = p, nbr
           if(.not.qrm_allocated(a%blocks(l,p)%c)) exit
        
           mm = min(a%f(l+1)-a%f(l), im-a%f(l)+1)           ! this block is a killer if:
           if(  l.eq.p        .or.                        & ! it's the diagonal block or
                (l-ki.ge.ibh)  .and.                      & ! it's at least bh below the last killer
                (mm.ge.kk) .and.                          & ! and it has more rows than columns
                .not. qrm_allocated(a%blocks(l,p)%stair) )& ! and does not have a staircase structure
             then

              nk = nk+1
              ks(nk)=l
              ki=l
           end if
        end do
        ks(nk+1)=l

        ! the list of killers is now ready. go through it
        ! write(*,'(" ")')
        do ki=nk, 1, -1
           i = ks(ki)


           if(i.ne.p) then
              do j=1, nbcx
                 nn = min(in-x%f(j)+1,x%f(j+1)-x%f(j))
                 ! write(*,'("tpmqrt t:",i2,x,i2,"->",i2,x,i2,"  kk:",i3," nn:",i3)')p,p,i,p,kk,nn
                 ! write(*,*)'col',p,'killing ',i, 'using',p,qrm_allocated(t%blocks(i,nbe+p)%c)
                 if(qrm_allocated(a%blocks(i,p)%c))        &
                      call sqrm_hitpmqrt_task(qrm_dscr,    &
                           qrm_no_transp,                  &
                           kk, nn, kk, kk, max(kk,nn), ib, &
                           1, 1,                           &
                           a%blocks(i,p),                  &
                           t%blocks(i,nbe+p),              &
                           x%blocks(p,j),                  &
                           x%blocks(i,j),                  &
                           work, prio)
              end do
           end if

           
           do l=ks(ki+1)-1, i+1, -1
              mm = min(im-a%f(l)+1,a%f(l+1)-a%f(l))
              ll = 0
              do j=1, nbcx                              
                 nn = min(in-x%f(j)+1,x%f(j+1)-x%f(j))
                 ! write(*,'("  tpmqrt s:",i2,x,i2,"->",i2,x,i2,"  mm:",i3," kk:",i3," nn:",i3)')i,p,l,p,mm,kk,nn
                 ! write(*,*)'col',p,'killing ',l, 'using',i
                 if(qrm_allocated(a%blocks(l,p)%c))         &
                      call sqrm_hitpmqrt_task(qrm_dscr,     &
                           qrm_no_transp,                   &
                           mm, nn, kk, ll, max(nn,kk), ib,  &
                           1, 1,                            &
                           a%blocks(l,p),                   &
                           t%blocks(l,p),                   &
                           x%blocks(i,j),                   &
                           x%blocks(l,j),                   &
                           work, prio)
              end do
           end do


           mm = min(a%f(i+1)-a%f(i), im-a%f(i)+1)
           do j=1, nbcx
              nn = min(in-x%f(j)+1,x%f(j+1)-x%f(j))
              ! write(*,'("geqrt  :",i2,x,i2,"  mm:",i3," kk:",i3," nn:",i3)')i,p,mm,kk,nn
              if(qrm_allocated(t%blocks(i,p)%c))       &
              call sqrm_higemqrt_task(qrm_dscr,        &
                   qrm_no_transp,                      &
                   mm, nn, kk, max(nn,kk), ib,         &
                   1, 1,                               &
                   a%blocks(i,p),                      &
                   t%blocks(i,p),                      &
                   x%blocks(i,j),                      &
                   work, prio)               
           end do                                       


           
        end do

     end do

  end if

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return

end subroutine sqrm_dsmat_gemqr_async




subroutine sqrm_dsmat_gemqr_async_old(qrm_dscr, trans, a, t, x, ib, bh, work, m, n, k, prio)

  use sqrm_dsmat_mod, alias=>sqrm_dsmat_gemqr_async
  use qrm_dscr_mod
  use sqrm_dsmat_task_mod

  implicit none

  type(qrm_dscr_type)         :: qrm_dscr
  character                   :: trans
  type(sqrm_dsmat_type)       :: a, t, x
  type(sqrm_ws_type)          :: work
  integer                     :: ib, bh
  integer, optional           :: m, n, k, prio

  integer                     :: i, j, p, s, l, nbr, nbc, nbcx, nbe, first, ibh
  integer                     :: mm, nn, kk, ll, im, in, ik

  ! error handling
  integer                     :: info
  character(len=*), parameter :: name='qrm_dsmat_gemqr_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  __QRM_IFPRESENT3(im, m, a%m)
  __QRM_IFPRESENT3(in, n, x%n)
  __QRM_IFPRESENT3(ik, k, a%n)

  ik = min(im,ik)

  if(min(ik, in).le.0) return
  
  nbr  = sqrm_dsmat_inblock(a, im) ! number of block-rows
  nbc  = sqrm_dsmat_inblock(a, ik) ! number of block-cols
  nbe  = min(nbr,nbc)
  nbcx = sqrm_dsmat_inblock(x, in)

  if(bh.le.0) then
     ibh = nbr
  else
     ibh = bh
  end if
  

  if(trans .eq. qrm_transp) then
     do p=1, nbe
        kk = min(ik-a%f(p)+1,a%f(p+1)-a%f(p))
        do first=p+1, nbr
           if(qrm_allocated(a%blocks(first,p)%stair)) exit
        end do
        do i = p, first-1, ibh
           mm = min(im-a%f(i)+1,a%f(i+1)-a%f(i))
           do j=1, nbcx
              nn = min(in-x%f(j)+1,x%f(j+1)-x%f(j))
              if(qrm_allocated(t%blocks(i,p)%c))       &
              call sqrm_higemqrt_task(qrm_dscr,        &
                   qrm_transp,                    &
                   mm, nn, kk, max(nn,kk), ib,         &
                   1, 1,                               &
                   a%blocks(i,p),                      &
                   t%blocks(i,p),                      &
                   x%blocks(i,j),                      &
                   work, prio)               
           end do                                       
           do l=i+1, min(i+ibh-1,first-1)                
              mm = min(im-a%f(l)+1,a%f(l+1)-a%f(l))
              ll = 0
              do j=1, nbcx                              
                 nn = min(in-x%f(j)+1,x%f(j+1)-x%f(j))
                 if(qrm_allocated(a%blocks(l,p)%c))    &
                 call sqrm_hitpmqrt_task(qrm_dscr,     &
                      qrm_transp,                 &
                      mm, nn, kk, ll, max(nn,kk), ib,  &
                      1, 1,                            &
                      a%blocks(l,p),                   &
                      t%blocks(l,p),                   &
                      x%blocks(i,j),                   &
                      x%blocks(l,j),                   &
                      work, prio)
              end do
           end do
        end do
        s = ibh
        do while (s.le.first-1-p+1)
           do i = p, first-1-s, 2*s
              l = i+s
              if(l.le.first-1) then
                 mm = min(im-a%f(l)+1,a%f(l+1)-a%f(l))
                 ll = mm
                 do j=1, nbcx
                    nn = min(in-x%f(j)+1,x%f(j+1)-x%f(j))
                    if(qrm_allocated(a%blocks(l,p)%c))   &
                    call sqrm_hitpmqrt_task(qrm_dscr,    &
                         qrm_transp,                &
                         mm, nn, kk, ll, max(nn,kk), ib, &
                         1, 1,                           &
                         a%blocks(l,p),                  &
                         t%blocks(l,nbe+p),              &
                         x%blocks(i,j),                  &
                         x%blocks(l,j),                  &
                         work, prio)
                 end do
              end if
           end do
           s = s*2
        end do

        do l=first, nbr
           mm = min(im-a%f(l)+1,a%f(l+1)-a%f(l))
           ll = 0
           do j=1, nbcx
              nn = min(in-x%f(j)+1,x%f(j+1)-x%f(j))
              if(qrm_allocated(a%blocks(l,p)%c))       &  
                   call sqrm_hitpmqrt_task(qrm_dscr,   &
                   qrm_transp,                    &
                   mm, nn, kk, ll, max(nn,kk), ib,     &
                   1, 1,                               &
                   a%blocks(l,p),                      &
                   t%blocks(l,p),                      &
                   x%blocks(p,j),                      &
                   x%blocks(l,j),                      &
                   work, prio)
           end do
        end do
     end do
  else
     do p=nbe, 1, -1
        kk = min(ik-a%f(p)+1,a%f(p+1)-a%f(p))
        do first=p+1, nbr
           if(qrm_allocated(a%blocks(first,p)%stair)) exit
        end do

        do l=nbr, first, -1
           mm = min(im-a%f(l)+1,a%f(l+1)-a%f(l))
           ll = 0
           do j=1, nbcx
              nn = min(in-x%f(j)+1,x%f(j+1)-x%f(j))
              if(qrm_allocated(a%blocks(l,p)%c))       &
                   call sqrm_hitpmqrt_task(qrm_dscr,   &
                   qrm_no_transp,                      &
                   mm, nn, kk, ll, max(nn,kk), ib,     &
                   1, 1,                               &
                   a%blocks(l,p),                      &
                   t%blocks(l,p),                      &
                   x%blocks(p,j),                      &
                   x%blocks(l,j),                      &
                   work, prio)
           end do
        end do

        s = ibh
        do while (s.le.first-1-p+1)
           s = s*2
        end do
        do while (s.ge.ibh)
           do i = p, first-1-s, 2*s
              l = i+s
              if(l.le.first-1) then
                 mm = min(im-a%f(l)+1,a%f(l+1)-a%f(l))
                 ll = mm
                 do j=1, nbcx
                    nn = min(in-x%f(j)+1,x%f(j+1)-x%f(j))
                    if(qrm_allocated(a%blocks(l,p)%c))      &
                         call sqrm_hitpmqrt_task(qrm_dscr,  &
                         qrm_no_transp,                     &
                         mm, nn, kk, ll, max(nn,kk), ib,    &
                         1, 1,                              &
                         a%blocks(l,p),                     &
                         t%blocks(l,nbe+p),                 &
                         x%blocks(i,j),                     &
                         x%blocks(l,j),                     &
                         work, prio)
                 end do
              end if
           end do
           s = s/2
        end do

        do i = p, first-1, ibh
           do l=min(i+ibh-1,first-1), i+1, -1
              mm = min(im-a%f(l)+1,a%f(l+1)-a%f(l))
              ll = 0
              do j=1, nbcx
                 nn = min(in-x%f(j)+1,x%f(j+1)-x%f(j))
                 if(qrm_allocated(a%blocks(l,p)%c))         &
                 call sqrm_hitpmqrt_task(qrm_dscr,          &
                      qrm_no_transp,                        &
                      mm, nn, kk, ll, max(nn,kk), ib,       &
                      1, 1,                                 &
                      a%blocks(l,p),                        &
                      t%blocks(l,p),                        &
                      x%blocks(i,j),                        &
                      x%blocks(l,j),                        &
                      work, prio)
              end do

           end do
           mm = min(im-a%f(i)+1,a%f(i+1)-a%f(i))
           do j=1, nbcx
              nn = min(in-x%f(j)+1,x%f(j+1)-x%f(j))
              if(qrm_allocated(t%blocks(i,p)%c))            &
              call sqrm_higemqrt_task(qrm_dscr,             &
                   qrm_no_transp,                           &
                   mm, nn, kk, max(nn,kk), ib,              &
                   1, 1,                                    &
                   a%blocks(i,p),                           &
                   t%blocks(i,p),                           &
                   x%blocks(i,j),                           &
                   work, prio)
           end do
        end do
     end do
  end if

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return

end subroutine sqrm_dsmat_gemqr_async_old
