!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT, Inria
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! computing a triangular system solution
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine sqrm_dsmat_potrf(uplo, a, work, m, k, prio, info)

  use sqrm_dsmat_mod, alias => sqrm_dsmat_potrf
  use qrm_dscr_mod
  use qrm_parameters_mod
  implicit none

  type(sqrm_dsmat_type)        :: a
  character                    :: uplo
  integer, optional            :: m, k, prio, info
  type(sqrm_ws_type), optional, target :: work

  type(qrm_dscr_type)          :: qrm_dscr
  type(sqrm_ws_type), pointer  :: iwork
  integer                      :: mb
  
  ! error handling
  integer                      :: err
  character(len=*), parameter  :: name='qrm_dsmat_potrf'

  err = 0

  call qrm_dscr_init(qrm_dscr)
  if(present(work)) then
     iwork => work
  else
     allocate(iwork)
     mb = maxval(a%f(2:a%nbc+1) - a%f(1:a%nbc))
     call qrm_ws_init(iwork, mb, mb)
  end if
  
  call sqrm_dsmat_potrf_async(qrm_dscr, uplo, a, iwork, m, k, prio)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, info)

  if(.not.present(work)) then
     call qrm_ws_destroy(iwork)
  end if
  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  return

end subroutine sqrm_dsmat_potrf

subroutine sqrm_dsmat_potrf_async(qrm_dscr, uplo, a, work, m, k, prio)

  use sqrm_dsmat_mod, alias => sqrm_dsmat_potrf_async
  use qrm_parameters_mod
  use qrm_dscr_mod
  use sqrm_dsmat_task_mod
  implicit none

  type(qrm_dscr_type)         :: qrm_dscr
  type(sqrm_dsmat_type)       :: a
  character                   :: uplo
  type(sqrm_ws_type)          :: work
  integer, optional           :: prio, m, k

  integer                     :: i, j, l
  integer                     :: im, in, ik, mm, nn, kk, last_mblock, last_kblock

  ! error handling
  integer                     :: info
  character(len=*), parameter :: name='qrm_dsmat_potrf_async'

  if(qrm_dscr%info.ne.0) return
  if(min(a%m,a%n).le.0) return
  info = 0

  __QRM_IFPRESENT3(im, m, a%n)
  __QRM_IFPRESENT3(ik, k, a%m)

  ik = min(im,ik)
  if(ik.le.0) return
  
  last_mblock = sqrm_dsmat_inblock(a, im)
  last_kblock = sqrm_dsmat_inblock(a, ik)
  
  if(uplo.eq.qrm_upper) then
     do l=1, last_kblock
        if(l.eq.last_kblock) then
           kk = ik-a%f(l)+1
           mm = min(a%f(l+1)-a%f(l), im-a%f(l)+1)
        else
           kk = a%f(l+1)-a%f(l)
           mm = a%f(l+1)-a%f(l)
        end if

        call sqrm_potrf_task(qrm_dscr,             &
             uplo,                                 & 
             mm, kk,                               &
             a%blocks(l,l),                        & 
             work,                                 &
             prio )

        do j = l+1, last_mblock
           if(j.eq.last_mblock) then
              nn = im-a%f(j)+1
           else
              nn = a%f(j+1)-a%f(j)
           end if
           call sqrm_trsm_task(qrm_dscr,           & 
                qrm_left, uplo,                    &
                qrm_transp, qrm_no_unit,      &
                kk, nn, mm,                        &
                qrm_sone, a%blocks(l,l),           &
                a%blocks(l,j), prio)
           
           do i = l+1, j-1
              ! write(*,*)'gemm ---> ',l,i,j,
              call sqrm_gemm_task(qrm_dscr,        &
                   qrm_transp, qrm_no_transp, &
                   a%f(i+1)-a%f(i), nn, kk,        &
                   qrm_smone,                      &
                   a%blocks(l,i), a%blocks(l,j),   &
                   qrm_sone, a%blocks(i,j), prio)
           end do
           call sqrm_herk_task(qrm_dscr,           &
                uplo, qrm_transp,             &
                nn, kk,                            &
                qrm_smone, a%blocks(l,j),          &
                qrm_sone,  a%blocks(j,j), prio)
        end do
     end do
  else
  end if

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return

end subroutine sqrm_dsmat_potrf_async

