!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT, Inria
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!! This file contains the module which holds the data types used during the factorization.
!!
!! ##############################################################################################


#include "qrm_common.h"

!> This module contains the definition of a dense matrix
!!
module sqrm_dsmat_mod
#if defined (have_starpu)
  use fstarpu_mod
#endif
  use qrm_error_mod
  use qrm_mem_mod
  use qrm_dscr_mod
  use qrm_parameters_mod


  !> This type defines a workspace type
  type sqrm_ws_type
     !> the coefficients
     real(r32), pointer              :: c(:,:) => null()
     logical                            :: seq=.false.
#if defined (have_starpu)
     type(c_ptr)                        :: hdl = c_null_ptr
#endif
  end type sqrm_ws_type


  !> This type defines a block type
  type sqrm_block_type
     !> the coefficients
     real(r32), pointer              :: c(:,:)=>null()
     ! real(r32), allocatable          :: c(:,:)                ! the coefficients
     integer, allocatable               :: stair(:)              ! the stair traversing the block
     logical                            :: partitioned=.false.   ! whether the block is in partitioned state
#if defined (have_starpu)
     !> the handle
     type(c_ptr)                        :: hdl = c_null_ptr      ! StarPU handle
     type(c_ptr), allocatable           :: shdls(:)              ! subhandles to handle repartitioning
#endif
  end type sqrm_block_type


  type sqrm_dsmat_type
     ! integer                            :: m, n, mb, nb, nbr, nbc
     integer                            :: m, n, nbr, nbc
     integer                            :: d              ! used to count the number of zeros along the diagonal after factorization FIXME: move
     integer, allocatable               :: f(:)           ! first index of every block
     type(sqrm_block_type), allocatable :: blocks(:,:)
     logical                            :: inited=.false. ! FIXME: see if needed
     logical                            :: pin=.false.
     logical                            :: seq=.false.
  end type sqrm_dsmat_type

  interface qrm_block_partition
     module procedure sqrm_block_partition1, sqrm_block_partition2
  end interface qrm_block_partition

  interface sqrm_block_partition
     module procedure sqrm_block_partition1, sqrm_block_partition2
  end interface sqrm_block_partition

  interface qrm_block_unpartition
     module procedure sqrm_block_unpartition1, sqrm_block_unpartition2
  end interface qrm_block_unpartition
  interface sqrm_block_unpartition
     module procedure sqrm_block_unpartition1, sqrm_block_unpartition2
  end interface sqrm_block_unpartition

  interface qrm_block_destroy
     module procedure sqrm_block_destroy
  end interface qrm_block_destroy

  interface qrm_dsmat_init
     module procedure sqrm_dsmat_init
  end interface qrm_dsmat_init

  interface qrm_dsmat_destroy
     module procedure sqrm_dsmat_destroy
  end interface qrm_dsmat_destroy

  interface qrm_dsmat_print
     module procedure sqrm_dsmat_print
  end interface qrm_dsmat_print

  interface qrm_dsmat_clean_stair
     module procedure sqrm_dsmat_clean_stair
  end interface qrm_dsmat_clean_stair

  interface qrm_block_print
     module procedure sqrm_block_print
  end interface qrm_block_print

  interface qrm_ws_init
     module procedure sqrm_ws_init
  end interface qrm_ws_init

  interface qrm_ws_destroy
     module procedure sqrm_ws_destroy
  end interface qrm_ws_destroy

  interface qrm_block_extadd
     module procedure sqrm_block_extadd
  end interface qrm_block_extadd

  ! interface qrm_block_extadd_byblock
  !    module procedure sqrm_block_extadd_byblock
  ! end interface qrm_block_extadd_byblock

  interface qrm_dsmat_inblock
     module procedure sqrm_dsmat_inblock
  end interface qrm_dsmat_inblock

  interface qrm_dsmat_copy
     subroutine sqrm_dsmat_copy(a, b, ia, ja, ib, jb, m, n, l, transp, prio, info)
       use qrm_dscr_mod
       import :: sqrm_dsmat_type
       type(sqrm_dsmat_type)          :: a
       type(sqrm_dsmat_type)          :: b
       integer, optional              :: ia, ja, ib, jb
       integer, optional              :: m, n, l
       character, optional            :: transp
       integer, optional              :: prio
       integer, optional              :: info
     end subroutine sqrm_dsmat_copy
  end interface qrm_dsmat_copy

  interface qrm_dsmat_copy_async
     subroutine sqrm_dsmat_copy_async(qrm_dscr, a, b, ia, ja, ib, jb, m, n, l, transp, prio)
       use qrm_dscr_mod
       import :: sqrm_dsmat_type
       type(qrm_dscr_type), target    :: qrm_dscr
       type(sqrm_dsmat_type)          :: a
       type(sqrm_dsmat_type)          :: b
       integer, optional              :: ia, ja, ib, jb
       integer, optional              :: m, n, l
       character, optional            :: transp
       integer, optional              :: prio
     end subroutine sqrm_dsmat_copy_async
  end interface qrm_dsmat_copy_async

  interface qrm_dsmat_axpy
     subroutine sqrm_dsmat_axpy(a, b, ia, ja, ib, jb, m, n, l, alpha, prio, info)
       use qrm_dscr_mod
       use qrm_parameters_mod
       import :: sqrm_dsmat_type
       type(sqrm_dsmat_type)          :: a
       type(sqrm_dsmat_type)          :: b
       integer, optional              :: ia, ja, ib, jb
       integer, optional              :: m, n, l
       real(r32), optional         :: alpha
       integer, optional              :: prio
       integer, optional              :: info
     end subroutine sqrm_dsmat_axpy
  end interface qrm_dsmat_axpy

  interface qrm_dsmat_axpy_async
     subroutine sqrm_dsmat_axpy_async(qrm_dscr, a, b, ia, ja, ib, jb, m, n, l, alpha, prio)
       use qrm_dscr_mod
       use qrm_parameters_mod
       import :: sqrm_dsmat_type
       type(qrm_dscr_type), target    :: qrm_dscr
       type(sqrm_dsmat_type)          :: a
       type(sqrm_dsmat_type)          :: b
       integer, optional              :: ia, ja, ib, jb
       integer, optional              :: m, n, l
       real(r32), optional         :: alpha
       integer, optional              :: prio
     end subroutine sqrm_dsmat_axpy_async
  end interface qrm_dsmat_axpy_async

  interface qrm_dsmat_extadd
     subroutine sqrm_dsmat_extadd(a, b, i, j, m, n, l, inout, op, rowmap, colmap, prio, info)
       use qrm_dscr_mod
       import :: sqrm_dsmat_type
       type(sqrm_dsmat_type)          :: a
       type(sqrm_dsmat_type)          :: b
       integer                        :: i, j, m, n, l
       character                      :: inout, op
       integer                        :: rowmap(:,:)
       integer, optional              :: colmap(:,:)
       integer, optional              :: prio
       integer, optional              :: info
     end subroutine sqrm_dsmat_extadd
  end interface qrm_dsmat_extadd

  interface qrm_dsmat_extadd_async
     subroutine sqrm_dsmat_extadd_async(qrm_dscr, a, b, i, j, m, n, l, inout, op, rowmap, colmap &
#if defined(have_starpu)
          , hdl &
#endif
          , prio &
          )
       use qrm_dscr_mod
       import :: sqrm_dsmat_type
       type(qrm_dscr_type), target    :: qrm_dscr
       type(sqrm_dsmat_type)          :: a
       type(sqrm_dsmat_type)          :: b
       integer                        :: i, j, m, n, l
       character                      :: inout, op
       integer                        :: rowmap(:,:)
       integer, optional              :: colmap(:,:)
#if defined(have_starpu)
       type(c_ptr), optional          :: hdl
#endif
       integer, optional              :: prio
     end subroutine sqrm_dsmat_extadd_async
  end interface qrm_dsmat_extadd_async

  interface qrm_dsmat_nrm
     subroutine sqrm_dsmat_nrm(qrm_dsmat, nrm, m, n, prio, info)
       use qrm_dscr_mod
       use qrm_parameters_mod
       import :: sqrm_dsmat_type
       type(sqrm_dsmat_type)          :: qrm_dsmat
       real(r32)                      :: nrm
       integer, optional              :: m, n
       integer, optional              :: prio
       integer, optional              :: info
     end subroutine sqrm_dsmat_nrm
  end interface qrm_dsmat_nrm

  interface qrm_dsmat_nrm_async
     subroutine sqrm_dsmat_nrm_async(qrm_dscr, qrm_dsmat, nrm, m, n, prio)
       use qrm_dscr_mod
       use qrm_parameters_mod
       import :: sqrm_dsmat_type
       type(qrm_dscr_type)            :: qrm_dscr
       type(sqrm_dsmat_type)          :: qrm_dsmat
       real(r32)                      :: nrm
       integer, optional              :: m, n
       integer, optional              :: prio
     end subroutine sqrm_dsmat_nrm_async
  end interface qrm_dsmat_nrm_async


  interface qrm_dsmat_init_t
     subroutine sqrm_dsmat_init_t(qrm_dsmat_a, qrm_dsmat_t, ib, bh, info)
       import :: sqrm_dsmat_type
       type(sqrm_dsmat_type), intent(in)            :: qrm_dsmat_a
       type(sqrm_dsmat_type), intent(inout), target :: qrm_dsmat_t
       integer, intent(in)                          :: ib, bh
       integer, optional                            :: info
     end subroutine sqrm_dsmat_init_t
  end interface qrm_dsmat_init_t

  interface qrm_dsmat_tpqr
     subroutine sqrm_dsmat_tpqr(a, b, t, ib, work, m, n, l, prio, info)
       import :: sqrm_dsmat_type, sqrm_ws_type
       type(sqrm_dsmat_type)       :: a, b, t
       type(sqrm_ws_type)          :: work
       integer                     :: ib
       integer, optional           :: m, n, l
       integer, optional           :: prio
       integer, optional           :: info
     end subroutine sqrm_dsmat_tpqr
  end interface qrm_dsmat_tpqr

  interface qrm_dsmat_tpqr_async
     subroutine sqrm_dsmat_tpqr_async(qrm_dscr, a, b, t, ib, work, m, n, l, prio)
       use qrm_dscr_mod
       import :: sqrm_dsmat_type, sqrm_ws_type
       type(qrm_dscr_type)         :: qrm_dscr
       type(sqrm_dsmat_type)       :: a, b, t
       type(sqrm_ws_type)          :: work
       integer                     :: ib
       integer, optional           :: m, n, l, prio
     end subroutine sqrm_dsmat_tpqr_async
  end interface qrm_dsmat_tpqr_async

  interface qrm_dsmat_tpmqr
     subroutine sqrm_dsmat_tpmqr(v, a, b, t, ib, work, m, n, k, l, prio, info)
       import :: sqrm_dsmat_type, sqrm_ws_type
       type(sqrm_dsmat_type)       :: v, a, b, t
       type(sqrm_ws_type)          :: work
       integer                     :: ib
       integer, optional           :: m, n, k, l, prio
       integer, optional           :: info
     end subroutine sqrm_dsmat_tpmqr
  end interface qrm_dsmat_tpmqr

  interface qrm_dsmat_tpmqr_async
     subroutine sqrm_dsmat_tpmqr_async(qrm_dscr, v, a, b, t, ib, work, m, n, k, l, prio)
       use qrm_dscr_mod
       import :: sqrm_dsmat_type, sqrm_ws_type
       type(qrm_dscr_type)         :: qrm_dscr
       type(sqrm_dsmat_type)       :: v, a, b, t
       type(sqrm_ws_type)          :: work
       integer                     :: ib
       integer, optional           :: m, n, k, l, prio
     end subroutine sqrm_dsmat_tpmqr_async
  end interface qrm_dsmat_tpmqr_async


  interface qrm_dsmat_geqr
     subroutine sqrm_dsmat_geqr(a, t, ib, bh, work, m, n, prio, info)
       import :: sqrm_dsmat_type, sqrm_ws_type
       type(sqrm_dsmat_type)       :: a, t
       type(sqrm_ws_type)          :: work
       integer                     :: ib, bh
       integer, optional           :: m, n, prio
       integer, optional           :: info
     end subroutine sqrm_dsmat_geqr
  end interface qrm_dsmat_geqr

  interface qrm_dsmat_geqr_async
     subroutine sqrm_dsmat_geqr_async(qrm_dscr, a, t, ib, bh, work, m, n, prio)
       use qrm_dscr_mod
       import :: sqrm_dsmat_type, sqrm_ws_type
       type(qrm_dscr_type)         :: qrm_dscr
       type(sqrm_dsmat_type)       :: a, t
       type(sqrm_ws_type)          :: work
       integer                     :: ib, bh
       integer, optional           :: m, n, prio
     end subroutine sqrm_dsmat_geqr_async
  end interface qrm_dsmat_geqr_async

  interface qrm_dsmat_gemqr
     subroutine sqrm_dsmat_gemqr(trans, a, t, x, ib, bh, work, m, n, k, prio, info)
       use qrm_dscr_mod
       import :: sqrm_dsmat_type, sqrm_ws_type
       character                   :: trans
       type(sqrm_dsmat_type)       :: a, t, x
       type(sqrm_ws_type)          :: work
       integer                     :: ib, bh
       integer, optional           :: m, n, k, prio
       integer, optional           :: info
     end subroutine sqrm_dsmat_gemqr
  end interface qrm_dsmat_gemqr

  interface qrm_dsmat_gemqr_async
     subroutine sqrm_dsmat_gemqr_async(qrm_dscr, trans, a, t, x, ib, bh, work, m, n, k, prio)
       use qrm_dscr_mod
       import :: sqrm_dsmat_type, sqrm_ws_type
       type(qrm_dscr_type)         :: qrm_dscr
       character                   :: trans
       type(sqrm_dsmat_type)       :: a, t, x
       type(sqrm_ws_type)          :: work
       integer                     :: ib, bh
       integer, optional           :: m, n, k, prio
     end subroutine sqrm_dsmat_gemqr_async
  end interface qrm_dsmat_gemqr_async

  interface qrm_dsmat_trsm
     subroutine sqrm_dsmat_trsm(side, uplo, trans, diag, alpha, a, b, m, n, k, prio, info)
       use qrm_dscr_mod
       use qrm_parameters_mod
       import :: sqrm_dsmat_type
       type(sqrm_dsmat_type)       :: a, b
       real(r32)                :: alpha
       character                   :: side, uplo, trans, diag
       integer, optional           :: m, n, k, prio
       integer, optional           :: info
     end subroutine sqrm_dsmat_trsm
  end interface qrm_dsmat_trsm

  interface qrm_dsmat_trsm_async
     subroutine sqrm_dsmat_trsm_async(qrm_dscr, side, uplo, trans, diag, alpha, a, b, m, n, k, prio)
       use qrm_parameters_mod
       use qrm_dscr_mod
       import :: sqrm_dsmat_type
       type(qrm_dscr_type)         :: qrm_dscr
       type(sqrm_dsmat_type)       :: a, b
       real(r32)                :: alpha
       character                   :: side, uplo, trans, diag
       integer, optional           :: m, n, k, prio
     end subroutine sqrm_dsmat_trsm_async
  end interface qrm_dsmat_trsm_async
  
  interface qrm_dsmat_trmm
     subroutine sqrm_dsmat_trmm(side, uplo, trans, diag, alpha, a, b, m, n, k, prio, info)
       use qrm_dscr_mod
       use qrm_parameters_mod
       import :: sqrm_dsmat_type
       type(sqrm_dsmat_type)       :: a, b
       real(r32)                :: alpha
       character                   :: side, uplo, trans, diag
       integer, optional           :: m, n, k, prio
       integer, optional           :: info
     end subroutine sqrm_dsmat_trmm
  end interface qrm_dsmat_trmm

  interface qrm_dsmat_trmm_async
     subroutine sqrm_dsmat_trmm_async(qrm_dscr, side, uplo, trans, diag, alpha, a, b, m, n, k, prio)
       use qrm_parameters_mod
       use qrm_dscr_mod
       import :: sqrm_dsmat_type
       type(qrm_dscr_type)         :: qrm_dscr
       type(sqrm_dsmat_type)       :: a, b
       real(r32)                :: alpha
       character                   :: side, uplo, trans, diag
       integer, optional           :: m, n, k, prio
     end subroutine sqrm_dsmat_trmm_async
  end interface qrm_dsmat_trmm_async

  interface qrm_dsmat_gemm
     subroutine sqrm_dsmat_gemm(transa, transb, alpha, a, b, beta, c, m, n, k, prio, info)
       use qrm_parameters_mod
       use qrm_dscr_mod
       import :: sqrm_dsmat_type
       type(sqrm_dsmat_type)       :: a, b, c
       real(r32)                :: alpha, beta
       character                   :: transa, transb
       integer, optional           :: m, n, k, prio
       integer, optional           :: info
     end subroutine sqrm_dsmat_gemm
  end interface qrm_dsmat_gemm

  interface qrm_dsmat_gemm_async
     subroutine sqrm_dsmat_gemm_async(qrm_dscr, transa, transb, alpha, a, b, beta, c, m, n, k, prio)
       use qrm_parameters_mod
       use qrm_dscr_mod
       import :: sqrm_dsmat_type
       type(qrm_dscr_type)            :: qrm_dscr
       type(sqrm_dsmat_type), target  :: a, b, c
       real(r32)                   :: alpha, beta
       character                      :: transa, transb
       integer, optional              :: m, n, k, prio
     end subroutine sqrm_dsmat_gemm_async
  end interface qrm_dsmat_gemm_async

  interface qrm_dsmat_trdcn
     subroutine sqrm_dsmat_trdcn(a, d, eps, n, info)
       use qrm_parameters_mod
       import :: sqrm_dsmat_type
       type(sqrm_dsmat_type)       :: a
       integer                     :: d
       integer, optional           :: n, info
       real(r32)                   :: eps
     end subroutine sqrm_dsmat_trdcn
  end interface qrm_dsmat_trdcn

  interface qrm_dsmat_trdcn_async
     subroutine sqrm_dsmat_trdcn_async(qrm_dscr, a, d, eps, n)
       use qrm_parameters_mod
       use qrm_dscr_mod
       import :: sqrm_dsmat_type
       type(qrm_dscr_type)         :: qrm_dscr
       type(sqrm_dsmat_type)       :: a
       integer                     :: d
       integer, optional           :: n
       real(r32)                   :: eps
     end subroutine sqrm_dsmat_trdcn_async
  end interface qrm_dsmat_trdcn_async


  interface qrm_dsmat_potrf
     subroutine sqrm_dsmat_potrf(uplo, a, work, m, k, prio, info)
       import :: sqrm_dsmat_type, sqrm_ws_type
       type(sqrm_dsmat_type)        :: a
       character                    :: uplo
       type(sqrm_ws_type), optional, target :: work
       integer, optional            :: m, k, prio, info
     end subroutine sqrm_dsmat_potrf
  end interface qrm_dsmat_potrf

  interface qrm_dsmat_potrf_async
     subroutine sqrm_dsmat_potrf_async(qrm_dscr, uplo, a, work, m, k, prio)
       use qrm_dscr_mod
       import :: sqrm_dsmat_type, sqrm_ws_type
       type(qrm_dscr_type)         :: qrm_dscr
       type(sqrm_dsmat_type)       :: a
       character                   :: uplo
       type(sqrm_ws_type)          :: work
       integer, optional           :: m, k, prio
     end subroutine sqrm_dsmat_potrf_async
  end interface qrm_dsmat_potrf_async

  interface qrm_dsmat_sytrf
     subroutine sqrm_dsmat_sytrf(uplo, a, m, k, prio, info)
       import :: sqrm_dsmat_type
       type(sqrm_dsmat_type)       :: a
       character                   :: uplo
       integer, optional           :: m, k, prio, info
     end subroutine sqrm_dsmat_sytrf
  end interface qrm_dsmat_sytrf

  interface qrm_dsmat_sytrf_async
     subroutine sqrm_dsmat_sytrf_async(qrm_dscr, uplo, a, m, k, prio)
       use qrm_dscr_mod
       import :: sqrm_dsmat_type
       type(qrm_dscr_type)         :: qrm_dscr
       type(sqrm_dsmat_type)       :: a
       character                   :: uplo
       integer, optional           :: m, k, prio
     end subroutine sqrm_dsmat_sytrf_async
  end interface qrm_dsmat_sytrf_async

  interface qrm_higeqrt
     subroutine sqrm_higeqrt(qrm_dscr, m, n, nb, ib, a, t, work, prio)
       use qrm_dscr_mod
       import :: sqrm_block_type, sqrm_ws_type
       type(qrm_dscr_type)   :: qrm_dscr
       type(sqrm_block_type) :: a, t
       integer               :: m, n, nb, ib
       type(sqrm_ws_type)    :: work
       integer, optional     :: prio
     end subroutine sqrm_higeqrt
  end interface qrm_higeqrt

  interface qrm_higemqrt
     subroutine sqrm_higemqrt(qrm_dscr, m, n, k, nb, ib, a, t, c, work, prio)
       use qrm_dscr_mod
       import :: sqrm_block_type, sqrm_ws_type
       type(qrm_dscr_type)   :: qrm_dscr
       type(sqrm_block_type) :: a, t, c
       type(sqrm_ws_type)    :: work
       integer               :: nb, ib
       integer, optional     :: prio
     end subroutine sqrm_higemqrt
  end interface qrm_higemqrt

  interface qrm_hitpqrt
     subroutine sqrm_hitpqrt(qrm_dscr, m, n, l, nb, ib, a1, a2, t, work, prio)
       use qrm_dscr_mod
       import :: sqrm_block_type, sqrm_ws_type
       type(qrm_dscr_type)   :: qrm_dscr
       type(sqrm_block_type) :: a1, a2, t
       type(sqrm_ws_type)    :: work
       integer               :: m, n, l, nb, ib
       integer, optional     :: prio
     end subroutine sqrm_hitpqrt
  end interface qrm_hitpqrt

  interface qrm_hitmpqrt
     subroutine sqrm_hitpmqrt(qrm_dscr, m, n, k, l, nb, ib, v, t, c1, c2, work, prio)
       use qrm_dscr_mod
       import :: sqrm_block_type, sqrm_ws_type
       type(qrm_dscr_type)   :: qrm_dscr
       type(sqrm_block_type) :: v, t, c1, c2
       type(sqrm_ws_type)    :: work
       integer               :: m, n, k, l, nb, ib
       integer, optional     :: prio
     end subroutine sqrm_hitpmqrt
  end interface qrm_hitmpqrt

  interface qrm_dsmat_fill
     subroutine sqrm_dsmat_fill(a, init, ivalue, i, j, m, n, l, prio, info)
       use qrm_parameters_mod
       import :: sqrm_dsmat_type
       type(sqrm_dsmat_type), target   :: a
       character                       :: init
       integer, optional               :: i, j, m, n, l, prio
       real(r32), optional          :: ivalue 
       integer, optional               :: info
     end subroutine sqrm_dsmat_fill
  end interface qrm_dsmat_fill

  interface qrm_dsmat_fill_async
     subroutine sqrm_dsmat_fill_async(qrm_dscr, a, init, ivalue, i, j, m, n, l, prio)
       use qrm_parameters_mod
       use qrm_dscr_mod
       import :: sqrm_dsmat_type
       type(qrm_dscr_type)            :: qrm_dscr
       type(sqrm_dsmat_type), target  :: a
       character                      :: init
       integer, optional              :: i, j, m, n, l, prio
       real(r32), optional         :: ivalue 
     end subroutine sqrm_dsmat_fill_async
  end interface qrm_dsmat_fill_async

  interface qrm_dsmat_addi
     subroutine sqrm_dsmat_addi(a, v, prio, info)
       use qrm_parameters_mod
       import :: sqrm_dsmat_type
       type(sqrm_dsmat_type), target   :: a
       integer, optional               :: prio
       real(r32), optional          :: v
       integer, optional               :: info
     end subroutine sqrm_dsmat_addi
  end interface qrm_dsmat_addi

  interface qrm_dsmat_addi_async
     subroutine sqrm_dsmat_addi_async(qrm_dscr, a, v, prio)
       use qrm_parameters_mod
       use qrm_dscr_mod
       import :: sqrm_dsmat_type
       type(qrm_dscr_type)            :: qrm_dscr
       type(sqrm_dsmat_type), target  :: a
       integer, optional              :: prio
       real(r32), optional         :: v 
     end subroutine sqrm_dsmat_addi_async
  end interface qrm_dsmat_addi_async

contains

  subroutine sqrm_block_destroy(b, pin, seq, async, info)

    type(sqrm_block_type) :: b
    logical, optional     :: pin, seq, async
    integer, optional     :: info

    logical               :: iseq, iasync
    integer, target       :: nparts

    call qrm_dealloc(b%stair)
    
    if(.not.qrm_allocated(b%c)) return

    if(present(seq)) then
       iseq = seq
    else
       iseq = .false.
    end if

    __QRM_IFPRESENT3(iseq, seq, .false.)
    __QRM_IFPRESENT3(iasync, async, .false.)
    
#if defined (have_starpu)
    if(.not.iseq) then
       if(allocated(b%shdls)) then
          ! a partition was planned and thus it has to be cleaned up
          nparts = size(b%shdls)
          call fstarpu_data_partition_clean (b%hdl,nparts,b%shdls)
          deallocate(b%shdls)
       end if
       if(c_associated(b%hdl)) then
          if(iasync) then
             call fstarpu_data_unregister_submit(b%hdl)
          else             
             call fstarpu_data_unregister(b%hdl)
          end if
          b%hdl = c_null_ptr
       end if
    end if
#endif

    call qrm_dealloc(b%c, pin=pin)

    return
  end subroutine sqrm_block_destroy


  subroutine sqrm_block_unregister_async(b, info)

    type(sqrm_block_type) :: b
    integer, optional     :: info

    logical               :: iseq
    integer, target       :: nparts

#if defined (have_starpu)
    if(allocated(b%shdls)) then
       ! a partition was planned and thus it has to be cleaned up
       nparts = size(b%shdls)
       call fstarpu_data_partition_clean (b%hdl,nparts,b%shdls)
       deallocate(b%shdls)
    end if
    if(c_associated(b%hdl)) then
       call fstarpu_data_unregister_submit(b%hdl)
       b%hdl = c_null_ptr
    end if
#endif

    return
  end subroutine sqrm_block_unregister_async


  subroutine sqrm_dsmat_destroy(qrm_dsmat)

    type(sqrm_dsmat_type), target  :: qrm_dsmat

    integer :: i, j

    if(.not.qrm_dsmat%inited) return

    do j=1, size(qrm_dsmat%blocks,2)
       do i=1, size(qrm_dsmat%blocks,1)
          call sqrm_block_destroy(qrm_dsmat%blocks(i,j), pin=qrm_dsmat%pin, seq=qrm_dsmat%seq)
       end do
    end do

    if(allocated(qrm_dsmat%blocks)) deallocate(qrm_dsmat%blocks)

    if(qrm_allocated(qrm_dsmat%f)) call qrm_dealloc(qrm_dsmat%f)

    qrm_dsmat%inited = .false.

    return
  end subroutine sqrm_dsmat_destroy


  subroutine sqrm_ws_init(work, m, n, info)
    implicit none
    type(sqrm_ws_type), target :: work
    integer                    :: m, n
    integer, optional          :: info

    real(r32) :: x

#if defined (have_starpu)
    call fstarpu_matrix_data_register(work%hdl, -1, &
         c_null_ptr,   &
         m,            &
         m,            &
         n,            &
         c_sizeof(x))
#else
    call qrm_alloc(work%c, m, n, info=info)
#endif

    return
  end subroutine sqrm_ws_init


  subroutine sqrm_ws_destroy(w, info)

    type(sqrm_ws_type)    :: w
    integer, optional     :: info

#if defined (have_starpu)
    if(.not.w%seq) then
       if(c_associated(w%hdl)) then
          call fstarpu_data_unregister(w%hdl)
          w%hdl = c_null_ptr
       end if
    end if
#endif

    call qrm_dealloc(w%c)

    return
  end subroutine sqrm_ws_destroy

  subroutine sqrm_block_extadd(a, b, br, bc, i, j, m, n, l, inout, op, rowmap, colmap)
    use qrm_string_mod
    implicit none

    type(sqrm_dsmat_type)  :: a                    ! if copyout this is the
    ! matrix containing the source
    ! block. If copyin it contains
    ! the destinantion block

    type(sqrm_dsmat_type)  :: b                    ! if copyout contains the
    ! destination blocks; if
    ! copyin contains the source
    ! blocks
    integer                :: br, bc               ! block-row and block-column
    ! of the source/destination
    ! block in a
    integer                :: i, j, m, n, l        ! i and j are the first row
    ! and column within the (bc,br)
    ! block and m and n are the
    ! number of rows and cols to
    ! be copied. l is the number of
    ! rows in the upper trapezoidal
    ! part of the block (same meaning
    ! as in tpqrt
    character              :: inout, op
    integer                :: rowmap(:,:)          ! block-row mapping from a to b
    integer, optional      :: colmap(:,:)          ! block-col mapping from a to b


    integer                :: abr, abc ! block row and col in a (just a copy of br and bc)
    integer                :: bbr, bbc ! block row and col in b

    integer                :: ai, aj, agi, agj ! local (to block) and global row and col indices in a
    integer                :: bi, bj, bgj      ! local (to block) and global row and col indices in b

    integer                :: aroff, acoff ! row and column offsets in a

    if(.not. qrm_allocated(a%blocks(br,bc)%c)) return

    abr = br ! just a copy to avoid confusion with blocks ob b
    abc = bc ! just a copy to avoid confusion with blocks ob b

    aroff = a%f(abr)-1  ! row-offset in a
    acoff = a%f(abc)-1  ! col-offset in a

    ! This is very ugly duplicate code but this is done to minimize
    ! the "if"s in inner loops
    if(op.eq.'a') then
       if(inout.eq.'o') then
          if(present(colmap)) then
             colsaoc: do aj = j, j+n-1
                agj = acoff + aj                              ! global a col index of the current coefficient
                bbc = colmap(agj,2)                           ! block-col in b     of the current coefficient
                bj  = colmap(agj,3)                           ! col index within block-col bbc of b

                rowsaoc: do ai = i, min(i+aj-j+m-l,i+m-1)
                   agi = aroff + ai                           ! global a row index of the current coefficient
                   bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
                   bi  = rowmap(agi,3)                        ! row index within block-row bbr of b
                   b%blocks(bbr,bbc)%c(bi,bj) = b%blocks(bbr,bbc)%c(bi,bj)+a%blocks(abr,abc)%c(ai,aj)

                end do rowsaoc
             end do colsaoc
          else
             colsao: do aj = j, j+n-1
                agj = acoff + aj                                 ! global a col index of the current coefficient
                bgj = agj                                        ! global b col index of the current coefficient
                bbc = sqrm_dsmat_inblock(b, bgj)                 ! block-col in b     of the current coefficient
                bj  = bgj-b%f(bbc)+1                             ! col index within block-col bbc of b

                rowsao: do ai = i, min(i+aj-j+m-l,i+m-1)
                   agi = aroff + ai                           ! global a row index of the current coefficient
                   bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
                   bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

                   b%blocks(bbr,bbc)%c(bi,bj) = b%blocks(bbr,bbc)%c(bi,bj)+a%blocks(abr,abc)%c(ai,aj)

                end do rowsao
             end do colsao
          end if
       else if(inout.eq.'i') then
          if(present(colmap)) then
             colscaic: do aj = j, j+n-1
                agj = acoff + aj                               ! global a col index of the current coefficient
                bbc = colmap(agj,2)                           ! block-col in b     of the current coefficient
                bj  = colmap(agj,3)                           ! col index within block-col bbc of b

                rowscaic: do ai = i, min(i+aj-j+m-l,i+m-1)
                   agi = aroff + ai                            ! global a row index of the current coefficient
                   bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
                   bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

                   a%blocks(abr,abc)%c(ai,aj) = a%blocks(abr,abc)%c(ai,aj) + b%blocks(bbr,bbc)%c(bi,bj)

                end do rowscaic
             end do colscaic
          else
             colsai: do aj = j, j+n-1
                agj = acoff + aj                               ! global a col index of the current coefficient
                bgj = agj                                      ! global b col index of the current coefficient
                bbc = sqrm_dsmat_inblock(b, bgj) !(bgj-1)/b%mb+1 ! block-col in b     of the current coefficient
                bj  = bgj-b%f(bbc)+1 !mod(bgj-1,b%mb) + 1        ! col index within block-col bbc of b

                rowsai: do ai = i, min(i+aj-j+m-l,i+m-1)
                   agi = aroff + ai                            ! global a row index of the current coefficient
                   bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
                   bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

                   a%blocks(abr,abc)%c(ai,aj) = a%blocks(abr,abc)%c(ai,aj) + b%blocks(bbr,bbc)%c(bi,bj)

                end do rowsai
             end do colsai
          end if
       end if

    else if(op.eq.'c') then
       if(inout.eq.'o') then
          if(present(colmap)) then
             colscoc: do aj = j, j+n-1
                agj = acoff + aj                               ! global a col index of the current coefficient
                bbc = colmap(agj,2)                           ! block-col in b     of the current coefficient
                bj  = colmap(agj,3)                           ! col index within block-col bbc of b

                rowscoc: do ai = i, min(i+aj-j+m-l,i+m-1)
                   agi = aroff + ai                            ! global a row index of the current coefficient
                   bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
                   bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

                   b%blocks(bbr,bbc)%c(bi,bj) = a%blocks(abr,abc)%c(ai,aj)

                end do rowscoc
             end do colscoc
          else
             colsco: do aj = j, j+n-1
                agj = acoff + aj                                 ! global a col index of the current coefficient
                bgj = agj                                        ! global b col index of the current coefficient
                bbc = sqrm_dsmat_inblock(b, bgj) !(bgj-1)/b%mb+1 ! block-col in b     of the current coefficient
                bj  = bgj-b%f(bbc)+1 !mod(bgj-1,b%mb) + 1        ! col index within block-col bbc of b

                rowsco: do ai = i, min(i+aj-j+m-l,i+m-1)
                   agi = aroff + ai                            ! global a row index of the current coefficient
                   bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
                   bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

                   b%blocks(bbr,bbc)%c(bi,bj) = a%blocks(abr,abc)%c(ai,aj)

                end do rowsco
             end do colsco
          end if
       else if(inout.eq.'i') then
          if(present(colmap)) then
             colscic: do aj = j, j+n-1
                agj = acoff + aj                               ! global a col index of the current coefficient
                bbc = colmap(agj,2)                           ! block-col in b     of the current coefficient
                bj  = colmap(agj,3)                           ! col index within block-col bbc of b

                rowscic: do ai = i, min(i+aj-j+m-l,i+m-1)
                   agi = aroff + ai                            ! global a row index of the current coefficient
                   bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
                   bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

                   a%blocks(abr,abc)%c(ai,aj) = b%blocks(bbr,bbc)%c(bi,bj)

                end do rowscic
             end do colscic
          else
             colsci: do aj = j, j+n-1
                agj = acoff + aj                                 ! global a col index of the current coefficient
                bgj = agj                                        ! global b col index of the current coefficient
                bbc = sqrm_dsmat_inblock(b, bgj) !(bgj-1)/b%mb+1 ! block-col in b     of the current coefficient
                bj  = bgj-b%f(bbc)+1 !mod(bgj-1,b%mb) + 1        ! col index within block-col bbc of b

                rowsci: do ai = i, min(i+aj-j+m-l,i+m-1)
                   agi = aroff + ai                            ! global a row index of the current coefficient
                   bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
                   bi  = rowmap(agi,3)                        ! row index within block-row bbr of b
                   a%blocks(abr,abc)%c(ai,aj) = b%blocks(bbr,bbc)%c(bi,bj)

                end do rowsci
             end do colsci
          end if
       end if

    end if

    return
  end subroutine sqrm_block_extadd


  ! subroutine sqrm_block_extadd_byblock(a, b, abr, abc, f_bbr, f_bbc, l_bbr, l_bbc, i, j, m, n, l, inout, op, rowmap, colmap)
  !   use qrm_string_mod
  !   implicit none

  !   type(sqrm_dsmat_type)  :: a                    ! if copyout this is the
  !   ! matrix containing the source
  !   ! block. If copyin it contains
  !   ! the destinantion block

  !   type(sqrm_dsmat_type)  :: b                    ! if copyout contains the
  !   ! destination blocks; if
  !   ! copyin contains the source
  !   ! blocks
  !   integer                :: abr, abc             ! block-row and block-column
  !                                                  ! of the source/destination
  !                                                  ! block in a
  !   integer                :: bbr, bbc             ! block-row and block-column
  !                                                  ! of the destination/source
  !                                                  ! block in b
  !   integer                :: f_bbr, f_bbc, l_bbr, l_bbc     
  !   integer                :: i, j, m, n, l        ! i and j are the first row
  !                                                  ! and column within the (bc,br)
  !                                                  ! block and m and n are the
  !                                                  ! number of rows and cols to
  !                                                  ! be copied. l is the number of
  !                                                  ! rows in the upper trapezoidal
  !                                                  ! part of the block (same meaning
  !                                                  ! as in tpqrt
  !   character              :: inout, op
  !   integer                :: rowmap(:,:)          ! block-row mapping from a to b
  !   integer, optional      :: colmap(:,:)          ! block-col mapping from a to b


  !   ! integer                :: abr, abc ! block row and col in a (just a copy of br and bc)
  !   ! integer                :: bbr, bbc ! block row and col in b

  !   integer                :: ai, aj, agi, agj ! local (to block) and global row and col indices in a
  !   integer                :: bi, bj, bgj      ! local (to block) and global row and col indices in b

  !   integer                :: aroff, acoff ! row and column offsets in a

  !   if(.not. qrm_allocated(a%blocks(abr,abc)%c)) return

  !   aroff = a%f(abr)-1  ! row-offset in a
  !   acoff = a%f(abc)-1  ! col-offset in a

  !   ! This is very ugly duplicate code but this is done to minimize
  !   ! the "if"s in inner loops
  !   if(op.eq.'a') then
  !      if(inout.eq.'o') then
  !         if(present(colmap)) then
  !            colsaoc: do aj = j, j+n-1
  !               agj = acoff + aj                              ! global a col index of the current coefficient
  !               bbc = colmap(agj,2)                           ! block-col in b     of the current coefficient
  !               if(bbc.lt.f_bbc) cycle colsaoc
  !               if(bbc.gt.l_bbc) exit  colsaoc
  !               bj  = colmap(agj,3)                           ! col index within block-col bbc of b

  !               rowsaoc: do ai = i, min(i+aj-j+m-l,i+m-1)
  !                  agi = aroff + ai                           ! global a row index of the current coefficient
  !                  bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
  !                  if(bbr.lt.f_bbr) cycle rowsaoc
  !                  if(bbr.gt.l_bbr) exit  rowsaoc
  !                  bi  = rowmap(agi,3)                        ! row index within block-row bbr of b
  !                  b%blocks(bbr,bbc)%c(bi,bj) = b%blocks(bbr,bbc)%c(bi,bj)+a%blocks(abr,abc)%c(ai,aj)

  !               end do rowsaoc
  !            end do colsaoc
  !         else
  !            colsao: do aj = j, j+n-1
  !               agj = acoff + aj                                   ! global a col index of the current coefficient
  !               bgj = agj                                          ! global b col index of the current coefficient
  !               bbc = sqrm_dsmat_inblock(b, bgj)                   ! block-col in b     of the current coefficient
  !               if(bbc.lt.f_bbc) cycle colsao
  !               if(bbc.gt.l_bbc) exit  colsao
  !               bj  = bgj-b%f(bbc)+1                               ! col index within block-col bbc of b
                
  !               rowsao: do ai = i, min(i+aj-j+m-l,i+m-1)
  !                  agi = aroff + ai                                ! global a row index of the current coefficient
  !                  bbr = rowmap(agi,2)                             ! block-row in b     of the current coefficient
  !                  if(bbr.lt.f_bbr) cycle rowsao
  !                  if(bbr.gt.l_bbr) exit  rowsao
  !                  bi  = rowmap(agi,3)                             ! row index within block-row bbr of b

  !                  b%blocks(bbr,bbc)%c(bi,bj) = b%blocks(bbr,bbc)%c(bi,bj)+a%blocks(abr,abc)%c(ai,aj)

  !               end do rowsao
  !            end do colsao
  !         end if
  !      else if(inout.eq.'i') then
  !         if(present(colmap)) then
  !            colscaic: do aj = j, j+n-1
  !               agj = acoff + aj                              ! global a col index of the current coefficient
  !               bbc = colmap(agj,2)                           ! block-col in b     of the current coefficient
  !               if(bbc.lt.f_bbc) cycle colscaic
  !               if(bbc.gt.l_bbc) exit  colscaic
  !               bj  = colmap(agj,3)                           ! col index within block-col bbc of b

  !               rowscaic: do ai = i, min(i+aj-j+m-l,i+m-1)
  !                  agi = aroff + ai                           ! global a row index of the current coefficient
  !                  bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
  !                  if(bbr.lt.f_bbr) cycle rowscaic
  !                  if(bbr.gt.l_bbr) exit  rowscaic
  !                  bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

  !                  a%blocks(abr,abc)%c(ai,aj) = a%blocks(abr,abc)%c(ai,aj) + b%blocks(bbr,bbc)%c(bi,bj)

  !               end do rowscaic
  !            end do colscaic
  !         else
  !            colsai: do aj = j, j+n-1
  !               agj = acoff + aj                                   ! global a col index of the current coefficient
  !               bgj = agj                                          ! global b col index of the current coefficient
  !               bbc = sqrm_dsmat_inblock(b, bgj)                   ! block-col in b     of the current coefficient
  !               if(bbc.lt.f_bbc) cycle colsai
  !               if(bbc.gt.l_bbc) exit  colsai
  !               bj  = bgj-b%f(bbc)+1                               ! col index within block-col bbc of b

  !               rowsai: do ai = i, min(i+aj-j+m-l,i+m-1)
  !                  agi = aroff + ai                            ! global a row index of the current coefficient
  !                  bbr = rowmap(agi,2)                         ! block-row in b     of the current coefficient
  !                  if(bbr.lt.f_bbr) cycle rowsai
  !                  if(bbr.gt.l_bbr) exit  rowsai
  !                  bi  = rowmap(agi,3)                         ! row index within block-row bbr of b

  !                  a%blocks(abr,abc)%c(ai,aj) = a%blocks(abr,abc)%c(ai,aj) + b%blocks(bbr,bbc)%c(bi,bj)

  !               end do rowsai
  !            end do colsai
  !         end if
  !      end if

  !   else if(op.eq.'c') then
  !      if(inout.eq.'o') then
  !         if(present(colmap)) then
  !            colscoc: do aj = j, j+n-1
  !               agj = acoff + aj                              ! global a col index of the current coefficient
  !               bbc = colmap(agj,2)                           ! block-col in b     of the current coefficient
  !               if(bbc.lt.f_bbc) cycle colscoc
  !               if(bbc.gt.l_bbc) exit  colscoc
  !               bj  = colmap(agj,3)                           ! col index within block-col bbc of b

  !               rowscoc: do ai = i, min(i+aj-j+m-l,i+m-1)
  !                  agi = aroff + ai                           ! global a row index of the current coefficient
  !                  bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
  !                  if(bbr.lt.f_bbr) cycle rowscoc
  !                  if(bbr.gt.l_bbr) exit  rowscoc
  !                  bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

  !                  b%blocks(bbr,bbc)%c(bi,bj) = a%blocks(abr,abc)%c(ai,aj)

  !               end do rowscoc
  !            end do colscoc
  !         else
  !            colsco: do aj = j, j+n-1
  !               agj = acoff + aj                                 ! global a col index of the current coefficient
  !               bgj = agj                                        ! global b col index of the current coefficient
  !               bbc = sqrm_dsmat_inblock(b, bgj)                 ! block-col in b     of the current coefficient
  !               if(bbc.lt.f_bbc) cycle colsco
  !               if(bbc.gt.l_bbc) exit  colsco
  !               bj  = bgj-b%f(bbc)+1                             ! col index within block-col bbc of b

  !               rowsco: do ai = i, min(i+aj-j+m-l,i+m-1)
  !                  agi = aroff + ai                            ! global a row index of the current coefficient
  !                  bbr = rowmap(agi,2)                         ! block-row in b     of the current coefficient
  !                  if(bbr.lt.f_bbr) cycle rowsco
  !                  if(bbr.gt.l_bbr) exit  rowsco
  !                  bi  = rowmap(agi,3)                         ! row index within block-row bbr of b

  !                  b%blocks(bbr,bbc)%c(bi,bj) = a%blocks(abr,abc)%c(ai,aj)

  !               end do rowsco
  !            end do colsco
  !         end if
  !      else if(inout.eq.'i') then
  !         if(present(colmap)) then
  !            colscic: do aj = j, j+n-1
  !               agj = acoff + aj                              ! global a col index of the current coefficient
  !               bbc = colmap(agj,2)                           ! block-col in b     of the current coefficient
  !               if(bbc.lt.f_bbc) cycle colscic
  !               if(bbc.gt.l_bbc) exit  colscic
  !               bj  = colmap(agj,3)                           ! col index within block-col bbc of b

  !               rowscic: do ai = i, min(i+aj-j+m-l,i+m-1)
  !                  agi = aroff + ai                           ! global a row index of the current coefficient
  !                  bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
  !                  if(bbr.lt.f_bbr) cycle rowscic
  !                  if(bbr.gt.l_bbr) exit  rowscic
  !                  bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

  !                  a%blocks(abr,abc)%c(ai,aj) = b%blocks(bbr,bbc)%c(bi,bj)

  !               end do rowscic
  !            end do colscic
  !         else
  !            colsci: do aj = j, j+n-1
  !               agj = acoff + aj                                 ! global a col index of the current coefficient
  !               bgj = agj                                        ! global b col index of the current coefficient
  !               bbc = sqrm_dsmat_inblock(b, bgj)                 ! block-col in b     of the current coefficient
  !               if(bbc.lt.f_bbc) cycle colsci
  !               if(bbc.gt.l_bbc) exit  colsci
  !               bj  = bgj-b%f(bbc)+1                             ! col index within block-col bbc of b

  !               rowsci: do ai = i, min(i+aj-j+m-l,i+m-1)
  !                  agi = aroff + ai                              ! global a row index of the current coefficient
  !                  bbr = rowmap(agi,2)                           ! block-row in b     of the current coefficient
  !                  if(bbr.lt.f_bbr) cycle rowsci
  !                  if(bbr.gt.l_bbr) exit  rowsci
  !                  bi  = rowmap(agi,3)                           ! row index within block-row bbr of b
  !                  a%blocks(abr,abc)%c(ai,aj) = b%blocks(bbr,bbc)%c(bi,bj)

  !               end do rowsci
  !            end do colsci
  !         end if
  !      end if

  !   end if

  !   return
  ! end subroutine sqrm_block_extadd_byblock

  subroutine sqrm_block_partition1(a)
    implicit none

    type(sqrm_block_type) :: a
    integer               :: parts

    if(qrm_allocated(a%c)) then
       if(.not.a%partitioned) then
#if defined (have_starpu)
          if(allocated(a%shdls)) then
             parts = size(a%shdls)
             call fstarpu_data_partition_submit(a%hdl, parts, a%shdls)
          else
             return
          end if
#endif
          a%partitioned = .true.
       end if
    end if

    return
  end subroutine sqrm_block_partition1

  subroutine sqrm_block_partition2(a, b)
    implicit none

    type(sqrm_block_type) :: a, b
    integer               :: parts

    if(qrm_allocated(a%c)) then
       if(.not.a%partitioned) then
#if defined (have_starpu)
          if(allocated(a%shdls)) then
             parts = size(a%shdls)
             call fstarpu_data_partition_submit(a%hdl, parts, a%shdls)
          else
             return
          end if
#endif
          a%partitioned = .true.
       end if
    end if

    if(qrm_allocated(b%c)) then
       if(.not.b%partitioned) then
#if defined (have_starpu)
          if(allocated(b%shdls)) then
             parts = size(b%shdls)
             call fstarpu_data_partition_submit(b%hdl, parts, b%shdls)
          else
             return
          end if
#endif
          b%partitioned = .true.
       end if
    end if

    return
  end subroutine sqrm_block_partition2

  subroutine sqrm_block_unpartition1(a)
    implicit none

    type(sqrm_block_type) :: a
    integer               :: parts

    if(qrm_allocated(a%c)) then
       if(a%partitioned) then
#if defined (have_starpu)
          parts = size(a%shdls)
          call fstarpu_data_unpartition_submit(a%hdl, parts, a%shdls, 0)
#endif
          a%partitioned = .false.
       end if
    end if

    return
  end subroutine sqrm_block_unpartition1

  subroutine sqrm_block_unpartition2(a, b)
    implicit none

    type(sqrm_block_type) :: a, b
    integer               :: parts

    if(qrm_allocated(a%c)) then
       if(a%partitioned) then
#if defined (have_starpu)
          parts = size(a%shdls)
          call fstarpu_data_unpartition_submit(a%hdl, parts, a%shdls, 0)
#endif
          a%partitioned = .false.
       end if
    end if

    if(qrm_allocated(b%c)) then
       if(b%partitioned) then
#if defined (have_starpu)
          parts = size(b%shdls)
          call fstarpu_data_unpartition_submit(b%hdl, parts, b%shdls, 0)
#endif
          b%partitioned = .false.
       end if
    end if

    return
  end subroutine sqrm_block_unpartition2


  subroutine sqrm_block_print(qrm_block, fm, unit)
    implicit none

    type(sqrm_block_type)  :: qrm_block
    character(len=*)       :: fm
    integer, optional      :: unit

    integer :: i, j, iunit, l
    character(len=:), allocatable :: zfm, dfm

    if(present(unit)) then
       iunit = unit
    else
       iunit = 6
    end if

    l = len(fm)

    if(fm(1:1).eq.'f') then
       i = 2
    else if(fm(1:1).eq.'e') then
       if(fm(2:2).eq.'s') then
          i = 3
       else
          i = 2
       end if
    else
       write(*,'("Invalid format. Returning.")')
       return
    end if

    do j=i, l
       if(fm(j:j).eq.'.') exit
    end do
    ! if(j.eq.l+1) j=j-1

    zfm = "("//fm(i:j-1)//"x,x"//")"
    dfm = "("//fm//",x)"

#if defined (have_starpu)
    if (c_associated(qrm_block%hdl)) then
       call fstarpu_data_acquire(qrm_block%hdl, FSTARPU_R)
    end if
#endif    
    do i=1, size(qrm_block%c,1)
       do j=1, size(qrm_block%c,2)
          write(iunit,dfm, advance='no')qrm_block%c(i,j)
       end do
       write(iunit,'(" ")')
    end do
    
#if defined (have_starpu)
    if (c_associated(qrm_block%hdl)) then
       call fstarpu_data_release(qrm_block%hdl)
    end if
#endif    
    return
  end subroutine sqrm_block_print

  subroutine sqrm_dsmat_init(qrm_dsmat, m, n, mb, mbs, fcs, pin, stair, seq, uplo, info)
    use qrm_parameters_mod
    use sqrm_utils_mod
    use qrm_common_mod
#if defined (have_starpu)
    use fstarpu_mod
    use sqrm_starpu_mod
#endif
    implicit none

    type(sqrm_dsmat_type), target  :: qrm_dsmat
    integer                        :: m, n
    integer, optional              :: mb, info, stair(:)
    integer, optional              :: mbs(:), fcs(:)
    logical, optional              :: pin, seq
    character, optional            :: uplo

    integer                        :: nbr, nbc, i, j, k, is, js, fr, fc, ib, ie, cnt, imb
    integer                        :: b, bs, nb, mn, mx
    type(c_ptr)                    :: bc_filter
    character                      :: iuplo
    integer, allocatable           :: imbs(:)

    ! error handling
    integer                        :: err
    character(len=*), parameter    :: name='qrm_dsmat_init'

    err = 0

    __QRM_IFPRESENT3(qrm_dsmat%seq, seq, .false.)
    __QRM_IFPRESENT3(qrm_dsmat%pin, pin, .false.)
    __QRM_IFPRESENT3(iuplo, uplo, qrm_no)

    if(.not.qrm_allocated(qrm_dsmat%f)) then
       call qrm_build_fc(qrm_dsmat%f, m, n, qrm_dsmat%nbr, qrm_dsmat%nbc, &
                         mb=mb, mbs=mbs, fc_in=fcs)
    end if

    nbr = qrm_dsmat%nbr
    nbc = qrm_dsmat%nbc
    qrm_dsmat%m  = m
    qrm_dsmat%n  = n

#if defined (have_starpu)
    if(.not.qrm_dsmat%seq) then
       call sqrm_starpu_init_codelets()
    end if
#endif

    allocate(qrm_dsmat%blocks(nbr, nbc))

    cols: do j=1, nbc
       fc = qrm_dsmat%f(j)                     ! first index in the column
       js = min(qrm_dsmat%f(j+1) - fc, n-fc+1)  ! block col-size

       if(iuplo.eq.qrm_upper) then
          ib=1; ie=j
       else if(iuplo.eq.qrm_lower) then
          ib=j; ie=nbr
       else
          ib=1; ie=nbr
       end if

       rows: do i=ib, ie

          fr = qrm_dsmat%f(i)                     ! first index in the row
          is = min(qrm_dsmat%f(i+1) - fr, m-fr+1)  ! block col-size

          ! if we get below the stair move to the next col
          if(present(stair)) then
             ! if we're below the staircase, then move to the next column
             if(fr .gt. stair(fc+js-1)) cycle cols
             ! if this block is traversed by the staircase, allocate its stair
             if((stair(fc).lt.fr+is-1) .and. (stair(fc+js-1).ge.fr)) then
                call qrm_alloc(qrm_dsmat%blocks(i,j)%stair, js, err)
                do k=1, js
                   qrm_dsmat%blocks(i,j)%stair(k) = min(max(stair(fc+k-1)-fr+1, 0),is)
                end do
             end if
          end if

          ! allocate the block
          call qrm_alloc(qrm_dsmat%blocks(i,j)%c, is, js, err, qrm_dsmat%pin)
          __QRM_INFO_CHECK(err, name, 'qrm_alloc', 9999)

#if defined (have_starpu)
          if(.not.qrm_dsmat%seq) then
             ! register the block to starpu
             call fstarpu_matrix_data_register(qrm_dsmat%blocks(i,j)%hdl, 0,           &
                  c_loc(qrm_dsmat%blocks(i,j)%c),         &
                  size(qrm_dsmat%blocks(i,j)%c,1),        &
                  size(qrm_dsmat%blocks(i,j)%c,1),        &
                  size(qrm_dsmat%blocks(i,j)%c,2),        &
                  c_sizeof(qrm_dsmat%blocks(i,j)%c(1,1)))
          end if
#endif
          qrm_dsmat%blocks(i,j)%partitioned = .false.
       end do rows
    end do cols

    qrm_dsmat%inited = .true.

9999 continue
    if(present(info)) info = err
    return
  end subroutine sqrm_dsmat_init




  subroutine sqrm_dsmat_print(qrm_dsmat, fm, unit, zero, i, j, m, n, l, ir)
    use qrm_parameters_mod
    implicit none

    type(sqrm_dsmat_type)         :: qrm_dsmat
    character(len=*)              :: fm
    logical, optional             :: zero, ir
    integer, optional             :: unit, i, j, m, n, l

    logical                       :: izero, iir, isnzero
    integer                       :: ii, ij, im, in, il, br, bc, iunit, ln, k, h
    integer                       :: row, col, ib, jb
    integer                       :: first_rblock, first_cblock, last_rblock, last_cblock
    character(len=:), allocatable :: zfm, dfm
    character                     :: w
    
    __QRM_IFPRESENT3(iunit,unit,6)
    __QRM_IFPRESENT3(izero,zero,.false.)
    __QRM_IFPRESENT3(ii,i,1)
    __QRM_IFPRESENT3(ij,j,1)
    __QRM_IFPRESENT3(im,m,qrm_dsmat%m-ii+1)
    __QRM_IFPRESENT3(in,n,qrm_dsmat%n-ij+1)
    __QRM_IFPRESENT3(il,l,0)
    __QRM_IFPRESENT3(iir,ir,.false.)

    ln = len(fm)

    if(min(qrm_dsmat%m,qrm_dsmat%n).le.0) return

    if(fm(1:1).eq.'f') then
       k = 2
    else if(fm(1:1).eq.'e') then
       if(fm(2:2).eq.'s') then
          k = 3
       else
          k = 2
       end if
    else
       write(*,'("Invalid format. Returning.")')
       return
    end if

    do h=k, ln
       if(fm(h:h).eq.'.') exit
    end do
    ! if(j.eq.ln+1) j=j-1

#if defined(dprec) || defined (sprec)
    zfm = "("//fm(k:h-1)//"x,x"//")"
    dfm = "("//fm//",x)"
#elif defined(zprec) || defined (cprec)
    if(iir) then
       w   = 'r' ! real part first
       zfm = "("//fm(k:h-1)//"x,x"//")"
       dfm = "("//fm//",x)"
    else
       w   = 'n'
       zfm = "("//fm(k:h-1)//"x"//fm(k:h-1)//"x,4x"//")"
       dfm = "(""("","//fm//","","","//fm//","")"",x)"
    end if
#endif

#if defined (have_starpu)
    do row = 1, size(qrm_dsmat%blocks, 1)
       do col = 1, size(qrm_dsmat%blocks, 2)
          if (c_associated(qrm_dsmat%blocks(row,col)%hdl)) then
             call fstarpu_data_acquire(qrm_dsmat%blocks(row,col)%hdl, FSTARPU_R)
          end if
       end do
    end do
#endif


    first_rblock = sqrm_dsmat_inblock(qrm_dsmat, ii)
    first_cblock = sqrm_dsmat_inblock(qrm_dsmat, ij)
    last_rblock  = sqrm_dsmat_inblock(qrm_dsmat, ii+im-1)
    last_cblock  = sqrm_dsmat_inblock(qrm_dsmat, ij+in-1)

20  continue

    do br=first_rblock, last_rblock
       do row = qrm_dsmat%f(br), qrm_dsmat%f(br+1)-1
          if((row.lt.ii).or.(row.gt.ii+im-1)) cycle ! skip row if out of the bounds
          do bc=first_cblock, last_cblock
             do col = qrm_dsmat%f(bc), qrm_dsmat%f(bc+1)-1
                if((col.lt.ij).or.(col.gt.ij+in-1)) cycle ! skip col if out of the bounds

                isnzero = qrm_allocated(qrm_dsmat%blocks(br,bc)%c) .and. &
                     (((il.ge.0).and.(row.le.ii+im-il+col-ij)) .or.      &
                     ((il.lt.0).and.(col.le.ij+in+il+row-ii)))

                ib = row-qrm_dsmat%f(br)+1
                jb = col-qrm_dsmat%f(bc)+1

                ! if(qrm_allocated(qrm_dsmat%blocks(br,bc)%stair)) then
                   ! isnzero = isnzero .and. &
                        ! ib.le.qrm_dsmat%blocks(br,bc)%stair(jb)
                ! end if

                if(isnzero) then
#if defined(dprec) || defined (sprec)
                   write(iunit,dfm, advance='no')qrm_dsmat%blocks(br,bc)%c(ib,jb)
#elif defined(zprec) || defined (cprec)
                   if(iir.and.w.eq.'r') then
                      write(iunit,dfm, advance='no')real(qrm_dsmat%blocks(br,bc)%c(ib,jb))
                   else if(iir.and.w.eq.'i') then
                      write(iunit,dfm, advance='no')aimag(qrm_dsmat%blocks(br,bc)%c(ib,jb))
                   else
                      write(iunit,dfm, advance='no')qrm_dsmat%blocks(br,bc)%c(ib,jb)
                   end if
#endif
                else
                   if(izero) then
#if defined(dprec) || defined (sprec)
                      write(iunit,dfm, advance='no')qrm_szero
#elif defined(zprec) || defined (cprec)
                      if(iir) then
                         write(iunit,dfm, advance='no')qrm_szero
                      else
                         write(iunit,dfm, advance='no')qrm_szero, qrm_szero
                      end if
#endif
                   else
                      write(iunit,zfm, advance='no')
                   end if
                end if
             end do
          end do
          write(iunit,'(" ")')
       end do
    end do

#if defined(zprec) || defined (cprec)
    if(w.eq.'r') then
       write(iunit,'(30("-"))')
       w = 'i'
       goto 20
    end if
#endif

#if defined (have_starpu)
    do row = 1, size(qrm_dsmat%blocks, 1)
       do col = 1, size(qrm_dsmat%blocks, 2)
          if (c_associated(qrm_dsmat%blocks(row,col)%hdl)) then
             call fstarpu_data_release(qrm_dsmat%blocks(row,col)%hdl)
          end if
       end do
    end do
#endif

    return

  end subroutine sqrm_dsmat_print



  function sqrm_dsmat_inblock(qrm_dsmat, i) result(b)
    implicit none

    type(sqrm_dsmat_type)  :: qrm_dsmat
    integer                :: i, b

    integer                :: nb, k

    b = 0

    if(i.gt. max(qrm_dsmat%m,qrm_dsmat%n)) return

    nb = max(qrm_dsmat%nbr, qrm_dsmat%nbc)
    
    do b=1, nb
       if((i.ge. qrm_dsmat%f(b)) .and. (i.lt. qrm_dsmat%f(b+1))) exit
    end do

    return

  end function sqrm_dsmat_inblock

  ! Assume a matrix partitioned into blocks of variable size whose
  ! first indices are given in a%f. We want to compute what part of
  ! the (br,bc) block overlaps with the trapezoidal submatrix
  ! described by i, j, m, n, l. This is given by bi, bj, bm, bn ,bl.
  subroutine sqrm_dsmat_block_ijmnl(a, i, j, m, n, l, br, bc, bi, bj, bm, bn, bl)

    implicit none

    type(sqrm_dsmat_type)  :: a
    integer                :: i, j, m, n, l, br, bc, bi, bj, bm, bn, bl

    integer                :: ii, jj, t
    ! write(*,*)'===> ',br, bc
    ii = max(i,a%f(br))
    jj = max(j,a%f(bc))
    bn = min(a%f(bc+1),j+n)-max(j,a%f(bc))
    bm = min(a%f(br+1),i+m)-max(i,a%f(br))
    ! write(*,*)i,j,m,n,l,ii,jj,bm,bn
    if (l.gt.0) then
       t  = i+m-l+jj-j
       if(t.lt.ii) then
          jj = jj+(ii-t)
          bn = bn-(ii-t)
          bm = min(bm,bn)
          bl = bm
       else
          bl = max(0,min(a%f(br+1),i+m)-t)
       end if
    else if (l.lt.0) then
       t  = j+n+l+ii-i
       if(t.lt.jj) then
          ii = ii+(jj-t)
          bm = bm-(jj-t)
          bn = min(bm,bn)
          bl = -bn
       else
          bl = -max(0,min(a%f(bc+1),j+n)-t)
       end if
    else
       bl = 0
    end if
    bj = max(jj-a%f(bc)+1, 1)
    bi = max(ii-a%f(br)+1, 1)

    return
  end subroutine sqrm_dsmat_block_ijmnl


  subroutine sqrm_dsmat_clean_stair(qrm_dsmat)
    use qrm_parameters_mod
    implicit none

    type(sqrm_dsmat_type)         :: qrm_dsmat

    integer                       :: row, col, j

    do row = 1, qrm_dsmat%nbr
       do col = 1, qrm_dsmat%nbc
#if defined (have_starpu)
          if (c_associated(qrm_dsmat%blocks(row,col)%hdl)) then
             call fstarpu_data_acquire(qrm_dsmat%blocks(row,col)%hdl, FSTARPU_R)
          end if
#endif
          if(qrm_allocated(qrm_dsmat%blocks(row,col)%stair)) then
             do j=1, min(qrm_dsmat%f(col+1)-qrm_dsmat%f(col),qrm_dsmat%n-qrm_dsmat%f(col)+1)
                qrm_dsmat%blocks(row,col)%c(qrm_dsmat%blocks(row,col)%stair(j)+1:,j)=qrm_szero
             end do
          end if

#if defined (have_starpu)
          if (c_associated(qrm_dsmat%blocks(row,col)%hdl)) then
             call fstarpu_data_release(qrm_dsmat%blocks(row,col)%hdl)
          end if
#endif
       end do
    end do

    return

  end subroutine sqrm_dsmat_clean_stair


  
end module sqrm_dsmat_mod
