!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_spfct_potrs.F90
!! This file contains a subroutine for computing the least-squares solution of a problem
!!
!!
!! ##############################################################################################

#include "qrm_common.h"

!> @brief This routine computes the solution of a SPD problem

!> This routine computed the solution of a SPD system
!! Ax=b with multiple RHSs.
!!
!! @param[in]   qrm_spfct a qrm_spfct_type data which contains the
!!                        result of the factorization.
!!                        On output the original data will be unchanged.
!! @param[in,out]       b the RHSs. A 2D array of leading dimension qrm_spfct%m.
!!                        On output b is modified
!! @param[out]          x the solution, i.e., R'\R\b
!!
subroutine zqrm_spfct_potrs2d(qrm_spfct, b, x, info)
  use zqrm_spmat_mod
  use zqrm_spfct_mod
  use qrm_error_mod
  use zqrm_analysis_mod
  use zqrm_factorization_mod
  use zqrm_solve_mod
  use zqrm_sdata_mod
  use qrm_dscr_mod
  implicit none

  type(zqrm_spfct_type), intent(in)   :: qrm_spfct
  complex(r64), target, intent(inout) :: b(:,:)
  complex(r64), target, intent(out)   :: x(:,:)
  integer, optional                   :: info

  type(zqrm_sdata_type), allocatable  :: qrm_sdata_fwd(:), qrm_sdata_bwd(:)
  integer                             :: i, nb, nrhs, n, gputh
  type(qrm_dscr_type)                 :: qrm_dscr

  ! error management
  integer                            :: err
  character(len=*), parameter        :: name='qrm_spfct_potrs'

  err = 0

  __QRM_PRNT_DBG('("Entering the spfct_potrs")')

  if(qrm_spfct%n .ne. qrm_spfct%m) then
     err = 31
     call qrm_error_print(err, name,ied=(/qrm_spfct%m,qrm_spfct%n/))
     goto 9999
  end if

  ! blocking to deal with multiple rhs
  call qrm_get(qrm_spfct, 'qrm_rhsnb', nb)
  nrhs = size(b,2)
  if(nb.le.0) nb = nrhs

  ! init the descriptor
  call qrm_get('qrm_gputh', gputh)
  if(min(nrhs,nb).ge.gputh) then
     call qrm_dscr_init(qrm_dscr)
  else
     call qrm_dscr_init(qrm_dscr, nocuda=.true.)
  end if

  allocate(qrm_sdata_fwd((nrhs-1)/nb+1))
  allocate(qrm_sdata_bwd((nrhs-1)/nb+1))
  
  do i=1, (nrhs-1)/nb+1
     call zqrm_sdata_init(qrm_sdata_fwd(i), qrm_spfct, &
          x(:,(i-1)*nb+1:min(i*nb,nrhs)),              &  ! LHS
          b(:,(i-1)*nb+1:min(i*nb,nrhs)))                 ! RHS
     call zqrm_sdata_init(qrm_sdata_bwd(i), qrm_spfct, &
          x(:,(i-1)*nb+1:min(i*nb,nrhs)),              &  ! LHS
          x(:,(i-1)*nb+1:min(i*nb,nrhs)),              &  ! RHS
          qrm_sdata_fwd(i)%front_slv)                     ! use the same fronts as fw

     call zqrm_spfct_trsm_async(qrm_dscr, qrm_spfct, qrm_conj_transp, qrm_sdata_fwd(i))
     call zqrm_spfct_trsm_async(qrm_dscr, qrm_spfct, qrm_no_transp, qrm_sdata_bwd(i))
  end do

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, err)
  
  call qrm_dscr_destroy(qrm_dscr)

  do i=1, (nrhs-1)/nb+1
     call zqrm_sdata_destroy(qrm_sdata_fwd(i))
     call zqrm_sdata_destroy(qrm_sdata_bwd(i))
  end do

9999 continue
  if(present(info)) info = err
  return

end subroutine zqrm_spfct_potrs2d




!> @brief This routine computes the solution of a sparse SPD problem

!> This routine computed the solution of a sparse SPD system
!! Ax=b with a single RHS.
!! 
!! @param[in]   qrm_spfct a qrm_spfct_type data which contains the
!!                        result of the factorization.
!!                        On output the original data will be unchanged.
!! @param[in,out]       b the RHSs. A 1D array of leading dimension qrm_spfct%m. On output
!!                        it will be modified
!! @param[out]          x the solution, i.e., R'\R\b
!!
subroutine zqrm_spfct_potrs1d(qrm_spfct, b, x, info)
  use zqrm_spfct_mod
  use qrm_error_mod
  use zqrm_utils_mod
  use zqrm_methods_mod, savesym => zqrm_spfct_potrs1d
  implicit none

  type(zqrm_spfct_type), intent(in)   :: qrm_spfct
  complex(r64), target, intent(inout) :: b(:)
  complex(r64), target, intent(out)   :: x(:)
  integer, optional                   :: info

  complex(r64), pointer               :: pntb(:,:), pntx(:,:)
  integer                             :: m, n
  ! error management
  integer                             :: err
  character(len=*), parameter         :: name='qrm_spfct_potrs1d'

  m = size(b,1)
  pntb(1:m,1:1) => b(1:m)

  n = size(x,1)
  pntx(1:n,1:1) => x(1:n)

  call zqrm_spfct_potrs2d(qrm_spfct, pntb, pntx, info)

  return

end subroutine zqrm_spfct_potrs1d
