!! ##############################################################################################
!!
!! Copyright 2012-2024 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!! This file contains a module that defines the data type for storing
!! the results of the solve phase (not used at the moment)
!!
!! ##############################################################################################


!> This module contains the definition of all the data related to the
!! solution phase. 
!! 
module zqrm_sdata_mod
  use iso_c_binding
  use zqrm_fdata_mod
  use qrm_adata_mod
#if defined (have_starpu)
  use fstarpu_mod
#endif


  ! Type: zqrm_rhs_type
  !
  ! This type defines a data structure containing a rhs block
  !
  type zqrm_rhs_type
     type(zqrm_dsmat_type)    :: rhs  ! the small RHSs assembled on each front
     integer                  :: nr, nc
     integer                  :: small
#if defined (have_starpu)            
     type(c_ptr)              :: hdl=c_null_ptr  ! symbolic handle
#endif
  end type zqrm_rhs_type

  ! Type: zqrm_sdata_type
  !
  ! This type defines a data structure containing a rhs block
  !
  type zqrm_sdata_type
     complex(r64), pointer              :: lhs(:,:), rhs(:,:)
#if defined (have_starpu)            
     type(c_ptr)                        :: lhdl=c_null_ptr
     type(c_ptr)                        :: rhdl=c_null_ptr
#endif
     type(zqrm_rhs_type), pointer       :: front_slv(:)=>null()
     logical                            :: free=.true.
     type(zqrm_ws_type)                 :: work
  end type zqrm_sdata_type

  interface qrm_sdata_init
     module procedure zqrm_sdata_init2d, zqrm_sdata_init1d
  end interface qrm_sdata_init

  interface zqrm_sdata_init
     module procedure zqrm_sdata_init2d, zqrm_sdata_init1d
  end interface zqrm_sdata_init

  interface qrm_sdata_destroy
     module procedure zqrm_sdata_destroy
  end interface qrm_sdata_destroy

  interface qrm_sync_sdata
     module procedure  zqrm_sync_sdata
  end interface qrm_sync_sdata

  interface qrm_sync_data
     module procedure :: zqrm_sync_sdata
  end interface qrm_sync_data  
  
contains
  
  subroutine zqrm_sdata_init2d(qrm_sdata, qrm_spfct, lhs, rhs, front_slv)
    use zqrm_spfct_mod
    use qrm_mem_mod
    implicit none

    type(zqrm_sdata_type)                 :: qrm_sdata
    type(zqrm_spfct_type)                 :: qrm_spfct
    complex(r64), target                  :: lhs(:,:), rhs(:,:)
    type(zqrm_rhs_type), target, optional :: front_slv(:)

    integer               :: node
    
    qrm_sdata%lhs => lhs
    qrm_sdata%rhs => rhs

#if defined (have_starpu)
    call fstarpu_matrix_data_register(qrm_sdata%lhdl, 0, &
         c_loc(qrm_sdata%lhs),                           &
         size(qrm_sdata%lhs,1),                          &
         size(qrm_sdata%lhs,1),                          &
         size(qrm_sdata%lhs,2),                          &
         c_sizeof(qrm_sdata%lhs(1,1)))

    call fstarpu_matrix_data_register(qrm_sdata%rhdl, 0, &
         c_loc(qrm_sdata%rhs),                           &
         size(qrm_sdata%rhs,1),                          &
         size(qrm_sdata%rhs,1),                          &
         size(qrm_sdata%rhs,2),                          &
         c_sizeof(qrm_sdata%rhs(1,1)))
#endif

    if(associated(qrm_sdata%front_slv)) then
       deallocate(qrm_sdata%front_slv)
       qrm_sdata%front_slv => null()
    end if
    
    call zqrm_ws_destroy(qrm_sdata%work)

    ! if provided, use an existing fronts list to speed up solve
    if(present(front_slv)) then
       !TODO: check for size
       qrm_sdata%front_slv => front_slv
       qrm_sdata%free = .false.
    else
       allocate(qrm_sdata%front_slv(qrm_spfct%adata%nnodes))
       qrm_sdata%free = .true.
#if defined (have_starpu)
       do node=1, qrm_spfct%adata%nnodes
          qrm_sdata%front_slv(node)%hdl = c_null_ptr
          if(qrm_spfct%adata%small(node) .ge. 0) then
             if(.not.c_associated(qrm_sdata%front_slv(node)%hdl)) then
                call fstarpu_void_data_register(qrm_sdata%front_slv(node)%hdl)
             end if
          end if
       end do
#endif
    end if
    
    return
  end subroutine zqrm_sdata_init2d

  subroutine zqrm_sdata_init1d(qrm_sdata, qrm_spfct, lhs, rhs, front_slv)
    use zqrm_spfct_mod
    use zqrm_utils_mod
    implicit none

    type(zqrm_sdata_type)         :: qrm_sdata
    type(zqrm_spfct_type)         :: qrm_spfct
    complex(r64), target          :: lhs(:), rhs(:)
    type(zqrm_rhs_type), optional :: front_slv(:)

    complex(r64), pointer    :: lhs_p(:,:), rhs_p(:,:)
    integer                  :: m, n
    
    ! Turn the 1D pointer into a 2D one and call the 2D method below
    m = size(lhs,1)
    lhs_p(1:m,1:1) => lhs(1:m)

    n = size(rhs,1)
    rhs_p(1:n,1:1) => rhs(1:n)

    call zqrm_sdata_init2d(qrm_sdata, qrm_spfct, lhs_p, rhs_p, front_slv)

  end subroutine zqrm_sdata_init1d

  
  subroutine zqrm_sdata_destroy(qrm_sdata)
    use qrm_mem_mod
    implicit none

    type(zqrm_sdata_type) :: qrm_sdata
    integer               :: i

#if defined (have_starpu)
    call fstarpu_data_unregister(qrm_sdata%lhdl)
    call fstarpu_data_unregister(qrm_sdata%rhdl)
    qrm_sdata%lhdl = c_null_ptr
    qrm_sdata%rhdl = c_null_ptr
#endif

    if(associated(qrm_sdata%front_slv) .and. qrm_sdata%free) then
       do i=1, size(qrm_sdata%front_slv,1)
          call zqrm_dsmat_destroy(qrm_sdata%front_slv(i)%rhs)
          
#if defined (have_starpu)
          if(c_associated(qrm_sdata%front_slv(i)%hdl)) then
             call fstarpu_data_unregister(qrm_sdata%front_slv(i)%hdl)
             qrm_sdata%front_slv(i)%hdl = c_null_ptr
          end if
#endif
       end do
       deallocate(qrm_sdata%front_slv)
       qrm_sdata%front_slv => null()
    else
       qrm_sdata%front_slv => null()
    end if
    
    call zqrm_ws_destroy(qrm_sdata%work)
    
    nullify(qrm_sdata%lhs)
    nullify(qrm_sdata%rhs)

    return
  end subroutine zqrm_sdata_destroy

  subroutine zqrm_sync_sdata(qrm_sdata)
    use qrm_common_mod
    use qrm_error_mod
    implicit none
    
    type(zqrm_sdata_type), target  :: qrm_sdata
    
    integer :: inode, node
    
#if defined(have_starpu)
    if(associated(qrm_sdata%front_slv)) then
       do inode = 1, size(qrm_sdata%front_slv,1)
          if(c_associated(qrm_sdata%front_slv(inode)%hdl)) then
             call fstarpu_data_acquire(qrm_sdata%front_slv(inode)%hdl, FSTARPU_R)
             call fstarpu_data_release(qrm_sdata%front_slv(inode)%hdl)
          end if
       end do
    end if
#endif
    return
    
  end subroutine zqrm_sync_sdata
  

end module zqrm_sdata_mod
